# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Generate model tensor level quantization config."""

from collections.abc import Sequence
import copy
from typing import Any, Optional, Union
import warnings

from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import default_policy as policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_QuantTrans = qtyping.QuantTransformation
_OpName = qtyping.TFLOperationName


class ParamsGenerator:
  """Generate model tensor level quantization parameters."""

  def __init__(self, float_tflite: Union[str, bytes]):
    self.flatbuffer_model = tfl_flatbuffer_utils.read_model(float_tflite)

    if not tfl_flatbuffer_utils.is_float_model(self.flatbuffer_model):
      warnings.warn(
          'Input model is already partially quantized. Proceeding with'
          ' re-quantization; existing quantized tensors will remain unchanged'
          ' regardless of the selected recipe.'
      )

    self._check_tensor_names_are_unique()
    self.buffer_to_tensors: dict[int, list[Any]] = (
        tfl_flatbuffer_utils.buffer_to_tensors(self.flatbuffer_model)
    )
    self.model_quant_results: dict[str, qtyping.TensorTransformationParams] = {}
    self._tensor_quant_params_cache = common_utils.TensorQuantParamsCache()

  def generate_quantization_parameters(
      self,
      model_recipe_manager: recipe_manager.RecipeManager,
      model_qsvs: Optional[dict[str, qtyping.QSV]] = None,
  ) -> dict[str, qtyping.TensorTransformationParams]:
    """Generate the quantization parameters for the model.

    Args:
      model_recipe_manager: The recipe manager for the model.
      model_qsvs: Quantization statistics values (QSVs) for the model. This is
        obtained through calibration process.

    Returns:
      model_quant_results: The quantization parameters for tensors in the model.

    Raises:
      RuntimeError: If the calibration dataset is required but not provided.
    """
    if model_recipe_manager.need_calibration() and not model_qsvs:
      raise RuntimeError(
          'Model quantization statistics values (QSVs) are required for the'
          ' input recipe. This can be obtained by running calibration on sample'
          ' dataset.'
      )

    if model_qsvs is None:
      model_qsvs = {}

    skip_subgraphs = set()
    op_codes = self.flatbuffer_model.operatorCodes
    for sg_ind, subgraph in enumerate(self.flatbuffer_model.subgraphs):

      graph_info = qtyping.GraphInfo(
          subgraph.tensors, self.flatbuffer_model.buffers
      )
      # Add input/output operators to the subgraph.
      subgraph.operators += (
          tfl_flatbuffer_utils.get_subgraph_input_output_operators(subgraph)
      )
      for subgraph_op_id, op in enumerate(subgraph.operators):
        # Get the op key.
        if isinstance(op, qtyping.IOOperator):
          op_key = op.op_key
          subgraph_op_id = -1  # Virtual op, no real id.
        else:
          op_code = op_codes[op.opcodeIndex].builtinCode
          # Do not quantize unknown ops.
          if op_code not in tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME:
            op_quant_results = self._get_params_for_no_quant_op(
                subgraph_op_id, op, subgraph.tensors
            )
            self._update_model_quant_results(op_quant_results)
            continue
          op_key = tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME[op_code]

        # Step1: query the quantization_recipe to get op config.
        op_scope = tfl_flatbuffer_utils.get_op_scope(op, subgraph.tensors)
        algorithm_name, op_quant_config = (
            model_recipe_manager.get_quantization_configs(op_key, op_scope)
        )

        if sg_ind in skip_subgraphs or policy.is_non_quantizable_composite_op(
            op
        ):
          algorithm_name = algorithm_manager.AlgorithmName.NO_QUANTIZE

        if algorithm_name == algorithm_manager.AlgorithmName.NO_QUANTIZE:
          side_effect_subgraphs = (
              tfl_flatbuffer_utils.get_op_side_effect_subgraphs(op)
          )
          skip_subgraphs.update(side_effect_subgraphs)

          op_quant_results = self._get_params_for_no_quant_op(
              subgraph_op_id, op, subgraph.tensors
          )

        else:
          op_info = qtyping.OpInfo(op, op_key, subgraph_op_id, op_quant_config)
          # Step2: query algorithm_manager to get/call the related function.
          materialize_func = algorithm_manager.get_quantization_func(
              algorithm_name,
              op_key,
              qtyping.QuantizeMode.MATERIALIZE,
          )
          op_quant_results = materialize_func(
              op_info=op_info,
              graph_info=graph_info,
              tensor_name_to_qsv=model_qsvs,
              tensor_quant_params_cache=self._tensor_quant_params_cache,
          )
        # Step3: update the results.
        self._update_model_quant_results(op_quant_results)
    self._post_process_results()
    return self.model_quant_results

  def _check_tensor_names_are_unique(self):
    """Checks if the tensor names are unique in the model."""
    global_tensor_names = set()
    for subgraph in self.flatbuffer_model.subgraphs:
      for tensor in subgraph.tensors:
        tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
        if tensor_name in global_tensor_names:
          raise ValueError(
              'Tensor name %s is not unique in the model. Please check your'
              ' model and rename the tensor as ParamsGenerator assumes tensor'
              ' names are unique.' % tensor_name
          )
        global_tensor_names.add(tensor_name)

  def _post_process_results(self) -> None:
    """Post process the quantization results.

    Raises:
      RuntimeError: If the tensors sharing the same buffer have different
      quantization settings.
    """
    self._check_and_fix_buffer_sharing()

  def _update_model_quant_results(
      self,
      op_tensor_results: list[qtyping.TensorTransformationParams],
  ) -> None:
    """Update the op quantization results to the final output.

    Args:
      op_tensor_results: Tensor level quantization params for the op.

    Raises:
      RuntimeError: If the same tensor has multiple quantization configs.
    """

    for op_tensor_result in op_tensor_results:
      tensor_name = op_tensor_result.tensor_name
      if tensor_name not in self.model_quant_results:
        self.model_quant_results[tensor_name] = copy.copy(op_tensor_result)
      else:
        tensor_params = self.model_quant_results[tensor_name]
        # Set source op.
        if op_tensor_result.producer is not None:
          # Src params must be unique (a tensor can only be produced by one op).
          if tensor_params.producer is not None:
            raise RuntimeError(
                'Tensor %s received multiple quantization parameters from the'
                ' source op, which should not happen as every tensor should'
                ' have only one source op.' % tensor_name
            )
          tensor_params.producer = op_tensor_result.producer
        # Set target op, which can be multiple (a tensor can be consumed by
        # multiple ops).
        if op_tensor_result.consumers is not None:
          if tensor_params.consumers is None:
            tensor_params.consumers = op_tensor_result.consumers[:]
          else:
            tensor_params.consumers += op_tensor_result.consumers[:]
        self.model_quant_results[tensor_name] = tensor_params

  def _get_params_for_no_quant_op(
      self,
      subgraph_op_id: int,
      op: Any,
      subgraph_tensors: list[Any],
  ) -> list[qtyping.TensorTransformationParams]:
    """Get the quantization parameters for ops require no quantization.

    Args:
      subgraph_op_id: The op id in the subgraph.
      op: The op that needs to be parsed.
      subgraph_tensors: Tensors in the subgraph.

    Returns:
      Tensor level quantization params for the op.
    """

    def no_quant_tensor_params():
      return qtyping.OpToTensorParams(
          subgraph_op_id=subgraph_op_id,
          transformations=[_QuantTrans.NO_QUANTIZE],
      )

    tensor_params = []
    for input_tensor_idx in op.inputs:
      if input_tensor_idx != -1:
        tensor = subgraph_tensors[input_tensor_idx]
        input_tensor_params = qtyping.TensorTransformationParams(
            tensor_name=tfl_flatbuffer_utils.get_tensor_name(tensor),
            consumers=[no_quant_tensor_params()],
        )
        tensor_params.append(input_tensor_params)

    for output_tensor_idx in op.outputs:
      if output_tensor_idx != -1:
        tensor = subgraph_tensors[output_tensor_idx]
        output_tensor_params = qtyping.TensorTransformationParams(
            tensor_name=tfl_flatbuffer_utils.get_tensor_name(tensor),
            producer=no_quant_tensor_params(),
        )
        tensor_params.append(output_tensor_params)
    return tensor_params

  def _mark_tensors_requiring_buffer_duplication(
      self, buffers_to_duplicate: Sequence[int]
  ) -> None:
    """Mark tensors that require buffer duplication.

    Marking a tensor means adding a DUPLICATE_BUFFER transformation as the first
    transformation to be applied for each consumer of the tensor. Need to do
    that for each consumer to preserve a zero layer and not affect the
    horizontal optimization later in the transformation instructions generator.

    Marks all tensors within each of the provided buffers as requiring buffer
    duplication, except for the last tensor. The order of tensors is assumed to
    be the same during both the marking and transformation performer steps, as
    determined by `self.buffer_to_tensors`. This allows the final tensor to
    reuse the original buffer, as it is not marked for duplication.

    Args:
      buffers_to_duplicate: Indices of the buffers to duplicate.
    """
    for buffer_idx in buffers_to_duplicate:
      for tensor in self.buffer_to_tensors[buffer_idx][:-1]:
        tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
        for consumer_params in self.model_quant_results[tensor_name].consumers:
          consumer_params.transformations.insert(
              0, _QuantTrans.DUPLICATE_BUFFER
          )

  def _mark_tensors_requiring_tensor_duplication(
      self, tensor_names_to_duplicate
  ) -> None:
    """Mark tensors that require tensor duplication.

    Marking a tensor means adding a DUPLICATE_TENSOR transformation as the first
    transformation to be applied for each consumer of the tensor. Need to do
    that for each consumer to preserve a zero layer and not affect the
    horizontal optimization later in the transformation instructions generator.

    Args:
      tensor_names_to_duplicate: Names of tensors to duplicate.
    """
    for tensor_name in tensor_names_to_duplicate:
      for consumer_params in self.model_quant_results[tensor_name].consumers:
        consumer_params.transformations.insert(0, _QuantTrans.DUPLICATE_TENSOR)

  def _check_buffer_sharing_for_tensor(self, tensor: Any) -> bool:
    """Check buffer sharing for the tensor against itself.

    Args:
      tensor: The tensor to check.

    Returns:
      Whether the tensor has compatible quantization parameters.

    Raises:
      RuntimeError: If the tensor has incompatible quantization parameters
      and the buffer is not constant.
    """
    tensor_params = self.model_quant_results.get(
        tfl_flatbuffer_utils.get_tensor_name(tensor), None
    )
    if tensor_params is None:
      return True

    if _are_tensor_consumer_params_compatible(tensor_params):
      return True
    elif _is_constant_tensor(tensor, self.flatbuffer_model.buffers):
      return False
    else:
      error_msg = (
          f'The tensor {tensor.name} consumers do not have the same'
          ' quantization parameters. Please modify your quantization recipe to'
          ' make sure the two tensors have the same quantization settings.'
      )
      raise RuntimeError(error_msg)

  def _check_buffer_sharing_for_self_compatible_tensors(
      self, tensor1: Any, tensor2: Any
  ) -> bool:
    """Checks that self-compatible tensors have the same quantization params.

    Self-compatible means that all tensor's consumers have the same quantization
    parameters.

    Args:
      tensor1: The first tensor to check.
      tensor2: The second tensor to check.

    Returns:
      Whether the tensors have compatible quantization parameters.

    Raises:
      RuntimeError: If the tensors have incompatible quantization parameters
      and the buffer is not constant.
    """
    tensor1_params = self.model_quant_results.get(
        tfl_flatbuffer_utils.get_tensor_name(tensor1), None
    )
    tensor2_params = self.model_quant_results.get(
        tfl_flatbuffer_utils.get_tensor_name(tensor2), None
    )

    if tensor1_params is None or tensor2_params is None:
      return True

    if _are_self_compatible_tensors_compatible_to_each_other(
        tensor1_params, tensor2_params
    ):
      return True
    elif _is_constant_tensor(tensor1, self.flatbuffer_model.buffers):
      return False
    else:
      error_msg = (
          f'The tensors {tensor1.name} and {tensor2.name} do not have'
          ' the same quantization parameters even though they share the'
          ' same buffer. Please modify your quantization recipe to make'
          ' sure the two tensors have the same quantization settings.'
      )
      raise RuntimeError(error_msg)

  def _check_and_fix_buffer_sharing(self) -> None:
    """Check and fix tensor/buffer sharing issues when possible.

    This function checks if tensors sharing the same buffer have the same
    quantization settings. If not, when it's possible, it will fix it by marking
    such tensors or buffers to be duplicated. Otherwise, it will raise an error.

    Possible cases that can be fixed by duplication:
      1. A constant tensor recieves different quantization parameters from its
      consumers. In this case, the tensor is marked for duplication.
      2. Two or more tensors share the same constant buffer and have different
      quantization parameters. In this case, the buffer is marked for
      duplication.

    Raises:
      RuntimeError: If the tensors sharing the same buffer have different
        quantization settings and it can't be resolved by duplicating the
        buffer/tensor.
    """
    buffers_to_duplicate = []
    tensor_names_to_duplicate = []
    for buffer_idx, tensors in self.buffer_to_tensors.items():
      # TODO: b/458797890 - Investigate if skipping buffer_idx == 0 is a
      # correct fix, or if it just covers up a deeper issue. This is only
      # required when statically quantizing models that have already been
      # quantized dynamically.
      if not tensors or buffer_idx == 0:
        continue
      # Check if any of the tensors needs to be duplicated.
      for tensor in tensors:
        if not self._check_buffer_sharing_for_tensor(tensor):
          tensor_names_to_duplicate.append(
              tfl_flatbuffer_utils.get_tensor_name(tensor)
          )
      # Check if the buffer needs to be duplicated.
      tensor_1 = tensors[0]
      tensor_name_1 = tfl_flatbuffer_utils.get_tensor_name(tensor_1)
      if tensor_name_1 in tensor_names_to_duplicate:
        buffers_to_duplicate.append(buffer_idx)
        continue
      for tensor_2 in tensors[1:]:
        tensor_name_2 = tfl_flatbuffer_utils.get_tensor_name(tensor_2)
        if (
            tensor_name_2 in tensor_names_to_duplicate
            or not self._check_buffer_sharing_for_self_compatible_tensors(
                tensor_1, tensor_2
            )
        ):
          buffers_to_duplicate.append(buffer_idx)
          break

    # Fix the buffer sharing issues.
    self._mark_tensors_requiring_buffer_duplication(buffers_to_duplicate)
    self._mark_tensors_requiring_tensor_duplication(tensor_names_to_duplicate)


def _are_tensor_consumer_params_compatible(
    params: qtyping.TensorTransformationParams,
) -> bool:
  """Check if all tensor's consumers have the same quantization parameters."""
  if params.consumers is None or len(params.consumers) < 2:
    return True
  consumer_1 = params.consumers[0]
  for consumer in params.consumers[1:]:
    if not _compatible_tensor_params(consumer, consumer_1):
      return False
  return True


def _are_self_compatible_tensors_compatible_to_each_other(
    params1: qtyping.TensorTransformationParams,
    params2: qtyping.TensorTransformationParams,
) -> bool:
  """Check if two self compatible tensors are compatible to each other.

  Self compatible means that all tensor's consumers have the same quantization
  parameters.

  Args:
    params1: The first tensor transformation params.
    params2: The second tensor transformation params.

  Returns:
    Whether the two tensors are compatible to each other.
  """
  # Check the producer.
  if params1.producer is None or params2.producer is None:
    if params1.producer != params2.producer:
      return False
  elif not _compatible_tensor_params(params1.producer, params2.producer):
    return False

  # Check the consumers.
  if params1.consumers is None or params2.consumers is None:
    if params1.consumers != params2.consumers:
      return False
  else:
    # Since all consumer params within each tensor are the same, it's enough to
    # check only the first consumers.
    if not _compatible_tensor_params(
        params1.consumers[0], params2.consumers[0]
    ):
      return False
  return True


def _same_tensor_params_except_id(
    params1: qtyping.OpToTensorParams,
    params2: qtyping.OpToTensorParams,
) -> bool:
  """Check if two op to tensor params are the same except for subgraph_op_id."""
  return params1.transformations == params2.transformations and (
      params1.parameters == params2.parameters
      or params1.parameters is None
      and params2.parameters is None
  )


def _compatible_tensor_params(
    params1: qtyping.OpToTensorParams,
    params2: qtyping.OpToTensorParams,
) -> bool:
  """Check if two op to tensor params are compatible."""
  float_source_transformations = [
      _QuantTrans.ADD_QUANTIZE,
      _QuantTrans.NO_QUANTIZE,
      _QuantTrans.INSERT_HADAMARD_ROTATION,
      _QuantTrans.INSERT_DECOMPOSED_HADAMARD_ROTATION,
  ]
  quantized_source_transformations = [
      _QuantTrans.QUANTIZE_TENSOR,
      _QuantTrans.ADD_DEQUANTIZE,
  ]
  if _same_tensor_params_except_id(params1, params2):
    return True
  # We only need to check the first transformation because transformations are
  # applied in order, and as long as the one that's immediately after the tensor
  # is the same, it's compatible.
  if (
      params1.transformations[0] in float_source_transformations
      and params2.transformations[0] in float_source_transformations
  ):
    return True
  if (
      params1.transformations[0] in quantized_source_transformations
      and params2.transformations[0] in quantized_source_transformations
      and params1.parameters == params2.parameters
  ):
    return True
  return False


def _is_constant_tensor(tensor: Any, buffers: Sequence[Any]) -> bool:
  """Check if the tensor is a constant tensor."""
  return buffers[tensor.buffer].data is not None



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for calibrator."""

from collections.abc import Generator
import pathlib
from typing import Any

from absl.testing import absltest
import numpy as np

import os
from ai_edge_quantizer import calibrator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_ComputePrecision = qtyping.ComputePrecision
_AlgorithmName = recipe_manager.AlgorithmName

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("")
_TENSOR_QUANT_CONFIG = qtyping.TensorQuantizationConfig

TEST_MIN_VAL, TEST_MAX_VAL = -1, 1

_RNG = np.random.default_rng(66)


def _representative_dataset_gen(size=(1, 8), num_samples=10):
  for _ in range(num_samples):
    vals = np.random.rand(*size).astype(np.float32)
    vals[0][0], vals[0][1] = (
        TEST_MIN_VAL,
        TEST_MAX_VAL,
    )  # fix min/max for testing
    yield {"input_1": vals}


def _get_calibration_data(
    dataset_gen: Generator[dict[str, Any], Any, None],
) -> dict[str, Any]:
  calibration_samples = [sample for sample in dataset_gen]
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _add_default_int8xint8_integer_recipe(recipe_manager_object):
  recipe_manager_object.add_quantization_config(
      regex=".*",
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
      op_config=qtyping.OpQuantizationConfig(
          activation_tensor_config=_TENSOR_QUANT_CONFIG(
              num_bits=8, symmetric=False
          ),
          weight_tensor_config=_TENSOR_QUANT_CONFIG(num_bits=8, symmetric=True),
          compute_precision=_ComputePrecision.INTEGER,
      ),
  )


class CalibratorTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "tests/models/single_fc.tflite"
    )
    self._calibrator = calibrator.Calibrator(self._test_model_path)
    self._recipe_manager = recipe_manager.RecipeManager()
    dataset_gen = _representative_dataset_gen()
    self._representative_dataset = _get_calibration_data(dataset_gen)

  def test_calibrator_state_manipulation(self):
    # load/get qsvs
    sample_qsv = {"serving_default_input_1:0": {"min": -10, "max": 8}}
    self._calibrator.load_model_qsvs(sample_qsv)
    model_tensor_qsvs = self._calibrator.get_model_qsvs()
    self.assertLen(model_tensor_qsvs, 1)
    self.assertIn("serving_default_input_1:0", model_tensor_qsvs)  # input
    input_qsv = model_tensor_qsvs["serving_default_input_1:0"]
    self.assertEqual(input_qsv["min"], -10)
    self.assertEqual(input_qsv["max"], 8)

    # reset qsvs
    self._calibrator.reset_model_qsvs()
    model_tensor_qsvs = self._calibrator.get_model_qsvs()
    self.assertEmpty(model_tensor_qsvs)

  def test_calibrate_single_fc_success(self):
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    self._calibrator.calibrate(
        self._representative_dataset, self._recipe_manager
    )
    model_tensor_qsvs = self._calibrator.get_model_qsvs()

    self.assertLen(model_tensor_qsvs, 2)
    self.assertIn("serving_default_input_1:0", model_tensor_qsvs)  # input
    input_qsv = model_tensor_qsvs["serving_default_input_1:0"]
    self.assertSequenceAlmostEqual(
        input_qsv["min"].flatten(), [TEST_MIN_VAL], delta=1e-5
    )
    self.assertSequenceAlmostEqual(
        input_qsv["max"].flatten(), [TEST_MAX_VAL], delta=1e-5
    )
    self.assertIn("StatefulPartitionedCall:0", model_tensor_qsvs)  # output
    output_qsv = model_tensor_qsvs["StatefulPartitionedCall:0"]
    # Relu, only check the min
    self.assertSequenceAlmostEqual(output_qsv["min"].flatten(), [0])

  def test_calibration_cache_is_empty_when_off(self):
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    self.assertEmpty(self._calibrator.get_cached_output())
    self._calibrator.calibrate(
        self._representative_dataset, self._recipe_manager, cache_output=False
    )
    self.assertEmpty(self._calibrator.get_cached_output())

  def test_calibration_cache_when_on(self):
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    self.assertEmpty(self._calibrator.get_cached_output())
    self._calibrator.calibrate(
        self._representative_dataset, self._recipe_manager, cache_output=True
    )
    self.assertLen(self._calibrator.get_cached_output(), 10)

  def test_calibration_cache_is_empty_after_reset(self):
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    self._calibrator.calibrate(
        self._representative_dataset, self._recipe_manager, cache_output=True
    )
    self._calibrator.clear_cached_output()
    self.assertEmpty(self._calibrator.get_cached_output())

  def test_calibrate_unsupported_ops_success(self):
    # Many ops in the following model are not supported currently.
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/branching_conv_fc.tflite"
    )
    test_calibrator = calibrator.Calibrator(test_model_path)
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    dataset_gen = _representative_dataset_gen(size=(3, 4, 4, 1))
    test_calibrator.calibrate(
        _get_calibration_data(dataset_gen),
        self._recipe_manager,
        cache_output=True,
    )
    self.assertLen(test_calibrator.get_cached_output(), 10)

  def test_calibrate_reshape_with_empty_shape_success(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/reshape_with_empty_shape.tflite"
    )
    test_calibrator = calibrator.Calibrator(test_model_path)
    _add_default_int8xint8_integer_recipe(self._recipe_manager)
    calib_data = tfl_interpreter_utils.create_random_normal_input_data(
        test_model_path, num_samples=4
    )
    test_calibrator.calibrate(calib_data, self._recipe_manager)
    self.assertNotEmpty(test_calibrator.get_model_qsvs())

  def test_save_and_load_calibration_result(self):
    # Setup some QSV
    sample_qsv = {
        "serving_default_input_1:0": {
            "min": np.array([-10.0]),
            "max": np.array([8.0]),
        }
    }
    self._calibrator.load_model_qsvs(sample_qsv)

    # Save
    temp_file = self.create_tempfile().full_path
    self._calibrator.save_calibration_result(temp_file)

    # Reset
    self._calibrator.reset_model_qsvs()
    self.assertEmpty(self._calibrator.get_model_qsvs())

    # Load
    self._calibrator.load_model_qsvs(temp_file)

    # Verify
    model_tensor_qsvs = self._calibrator.get_model_qsvs()
    self.assertLen(model_tensor_qsvs, 1)
    self.assertIn("serving_default_input_1:0", model_tensor_qsvs)
    input_qsv = model_tensor_qsvs["serving_default_input_1:0"]
    self.assertSequenceAlmostEqual(input_qsv["min"], [-10.0])
    self.assertSequenceAlmostEqual(input_qsv["max"], [8.0])


class CalibratorAlreadyQuantizedModelTest(absltest.TestCase):

  def test_check_is_float_model_succeeds_when_model_is_float(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/conv_fc_mnist.tflite"
    )
    _ = calibrator.Calibrator(test_model_path)

  def test_check_is_quantized_model_succeeds_when_model_is_quantized(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/mnist_quantized.tflite"
    )
    _ = calibrator.Calibrator(test_model_path)


class CalibratorToyGemma2Test(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)

    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/toy_model_with_kv_cache_multi_signature.tflite"
    )

    self._toy_gemma2_calibration_dataset = {
        "signature_1": [{
            "cache_0": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            "cache_1": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            "positions": (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
            "tokens": (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
        }],
        "signature_2": [{
            "cache_0": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            "cache_1": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            "positions": (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
            "tokens": (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
        }],
    }

  def test_toy_gemma2_calibration_success(self):
    calib = calibrator.Calibrator(self._test_model_path)
    recipe_mngr = recipe_manager.RecipeManager()
    _add_default_int8xint8_integer_recipe(recipe_mngr)
    calib.calibrate(
        self._toy_gemma2_calibration_dataset,
        model_recipe_manager=recipe_mngr,
    )
    self.assertLen(calib.get_model_qsvs(), 202)


class CalibrationInterpreterTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(0)
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "tests/models/single_fc.tflite"
    )

  def test_initialization(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    self.assertIsInstance(interpreter, calibrator.CalibrationInterpreter)

  def test_calibration_mode(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    runner = interpreter.get_signature_runner()

    # Run inference which triggers calibration
    input_data = np.random.rand(1, 8).astype(np.float32)
    output = runner(input_1=input_data)

    # Check results
    qsvs = interpreter.get_calibration_results()
    self.assertNotEmpty(qsvs)

    # Verify input tensor qsv
    self.assertIn("serving_default_input_1:0", qsvs)
    self.assertIsNotNone(output)

  def test_inference_mode(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.INFERENCE
    )
    runner = interpreter.get_signature_runner()

    input_data = np.random.rand(1, 8).astype(np.float32)
    output = runner(input_1=input_data)

    with self.assertRaisesRegex(
        ValueError, "Calibration results are not available in INFERENCE mode."
    ):
      interpreter.get_calibration_results()
    self.assertIsNotNone(output)

  def test_save_calibration_result(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    runner = interpreter.get_signature_runner()
    input_data = np.random.rand(1, 8).astype(np.float32)
    runner(input_1=input_data)

    temp_file = self.create_tempfile().full_path
    interpreter.save_calibration_result(temp_file)

    # Verify file exists and has content
    with open(temp_file, "r") as f:
      content = f.read()
      self.assertNotEmpty(content)

  def test_get_signature_list(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    signatures = interpreter.get_signature_list()
    self.assertNotEmpty(signatures)
    self.assertIn("serving_default", signatures)

  def test_runner_details(self):
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    runner = interpreter.get_signature_runner()
    input_details = runner.get_input_details()
    output_details = runner.get_output_details()

    self.assertNotEmpty(input_details)
    self.assertNotEmpty(output_details)
    self.assertIn("input_1", input_details)

  def test_output_match_original_interpreter(self):
    # Run calibration interpreter
    interpreter = calibrator.CalibrationInterpreter(
        self._test_model_path, mode=calibrator.CalibrationMode.CALIBRATION
    )
    calib_runner = interpreter.get_signature_runner()
    input_data = np.random.rand(1, 8).astype(np.float32)
    calib_output = calib_runner(input_1=input_data)

    # Run original interpreter
    original_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._test_model_path
    )
    original_runner = original_interpreter.get_signature_runner()
    original_output = original_runner(input_1=input_data)

    # Compare
    self.assertEqual(calib_output.keys(), original_output.keys())
    for key in calib_output:
      np.testing.assert_array_equal(calib_output[key], original_output[key])


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Quantization Calibration."""

from collections.abc import Callable, Iterable
import copy
import enum
import json
from typing import Any, Union

from absl import logging
import numpy as np

import os
from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import default_policy as policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import calibration_utils
from ai_edge_quantizer.utils import progress_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils


class CalibrationMode(enum.Enum):
  INFERENCE = 1
  CALIBRATION = 2


_SignatureInput = dict[str, Any]  # input_argument_name -> tensor_value.
_SignatureOutput = dict[
    str, np.ndarray
]  # output_argument_name -> tensor_value.


class CalibrationInterpreter:
  """A TFL interpreter-like interface for model calibration.

  This is a wrapper around Calibrator that replaces the TFL Interpreter to
  enable calibration. If mode is CALIBRATION, it runs calibration, otherwise it
  acts as a regular TFL interpreter for inference. When in CALIBRATION mode,
  each invocation of a signature runner will update the calibration statistics
  in self._calibrator. Calibrator is needed in both modes because it contains
  tfl interpreter instance to run the model.
  """

  def __init__(
      self,
      model_path: str,
      mode: CalibrationMode = CalibrationMode.INFERENCE,
      qsv_update_func: Callable[
          [qtyping.QSV, qtyping.QSV],
          qtyping.QSV,
      ] = calibration_utils.moving_average_update,
  ):
    """Initializes the CalibrationInterpreter.

    Args:
      model_path: The path to the TFLite model.
      mode: The mode of the interpreter. If CALIBRATION, the interpreter will
        preserve all tensors for calibration purposes.
      qsv_update_func: The function to update the QSVs across calibration steps.
    """
    self._calibrator = Calibrator(
        model_path,
        interpreter_preserve_all_tensors=(mode == CalibrationMode.CALIBRATION),
        qsv_update_func=qsv_update_func,
    )
    self._mode = mode

  def get_signature_runner(self, signature_key: str | None = None):
    """Returns the signature runner."""
    return CalibrationSignatureRunner(
        self._calibrator, signature_key, self._mode
    )

  def get_calibration_results(self):
    """Returns the calibration results."""
    if self._mode == CalibrationMode.INFERENCE:
      raise ValueError(
          "Calibration results are not available in INFERENCE mode."
      )
    return self._calibrator.get_model_qsvs()

  def save_calibration_result(self, output_path: str):
    """Saves the calibration results."""
    if self._mode == CalibrationMode.INFERENCE:
      raise ValueError(
          "Calibration results are not available in INFERENCE mode."
      )
    self._calibrator.save_calibration_result(output_path)

  def get_signature_list(self) -> list[str]:
    """Returns the signature list."""
    return self._calibrator.get_signature_list()


class CalibrationSignatureRunner:
  """Wrapper around TFL signature runner to enable calibration."""

  def __init__(
      self,
      calibrator_obj: "Calibrator",
      signature_key: str | None = None,
      mode: CalibrationMode = CalibrationMode.INFERENCE,
      quantization_recipe: recipe_manager.ModelQuantizationRecipe = recipe.static_wi8_ai8(),
  ):
    """Initializes the CalibrationSignatureRunner.

    Args:
      calibrator_obj: The Calibrator instance.
      signature_key: The key of the signature to run. If None, the default
        signature is used.
      mode: The mode of the runner. If CALIBRATION, invoking the runner will
        update 'calibrator_obj' with new quantization statistics values. If
        INFERENCE, the runner behaves like a standard signature runner.
      quantization_recipe: The quantization recipe to use for calibration.
        Defaults to static_wi8_ai8.
    """
    self._calibrator = calibrator_obj
    self._signature_key = signature_key
    self._mode = mode
    self._recipe_manager = recipe_manager.RecipeManager()
    self._recipe_manager.load_quantization_recipe(quantization_recipe)
    self._signature_runner = (
        self._calibrator._tfl_interpreter.get_signature_runner(
            self._signature_key
        )
    )

  def __call__(self, **kwargs):
    if self._mode == CalibrationMode.INFERENCE:
      return self._signature_runner(**kwargs)
    self._calibrator.calibrate(
        calibration_dataset={self._signature_key: [kwargs]},
        model_recipe_manager=self._recipe_manager,
        cache_output=True,
    )
    outputs = self._calibrator.get_cached_output()
    assert len(outputs) == 1
    self._calibrator.clear_cached_output()
    return outputs[0]

  def get_input_details(self):
    """Returns the input details of the model."""
    return self._signature_runner.get_input_details()

  def get_output_details(self):
    """Returns the output details of the model."""
    return self._signature_runner.get_output_details()


class Calibrator:
  """Calibrator for TFLite model."""

  def __init__(
      self,
      float_tflite: Union[str, bytes],
      num_threads: int = 16,
      interpreter_preserve_all_tensors: bool = True,
      qsv_update_func: Callable[
          [qtyping.QSV, qtyping.QSV],
          qtyping.QSV,
      ] = calibration_utils.moving_average_update,
  ):
    """Initializes the Calibrator.

    Args:
      float_tflite: The path to the TFLite model or the model content as bytes.
      num_threads: The number of threads to use for the TFLite interpreter.
      interpreter_preserve_all_tensors: Whether to preserve all tensors in the
        TFLite interpreter. This must be True for calibration.
      qsv_update_func: The function used to update Quantization Statistics
        Values (QSVs) across different calibration steps.
    """
    self._qsv_update_func = qsv_update_func
    self._flatbuffer_model = tfl_flatbuffer_utils.read_model(float_tflite)
    self._tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        float_tflite,
        use_xnnpack=True,
        num_threads=num_threads,
        preserve_all_tensors=interpreter_preserve_all_tensors,
    )
    # Tensor name to tensor content.
    self._tensor_content_map: dict[str, Any] = {}
    # QSV of all the tensors in the model.
    self._model_qsvs: dict[str, qtyping.QSV] = {}
    # Cached output of the model.
    self._cached_output: list[_SignatureOutput] = []

  def _get_total_operations(self, calibration_dataset) -> int:
    """Get the total number of OPs to go through while calibrating the model."""
    data_sizes = {
        key: len(list(dataset)) for key, dataset in calibration_dataset.items()
    }
    total_ops = 0
    for key, value in data_sizes.items():
      subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
          self._tfl_interpreter, key
      )
      subgraph_inds = [subgraph_idx]
      total_operators = 0
      while subgraph_inds:
        subgraph_idx = subgraph_inds.pop()
        subgraph = self._flatbuffer_model.subgraphs[subgraph_idx]
        subgraph_operators = subgraph.operators
        if not any(
            isinstance(op, qtyping.IOOperator) for op in subgraph.operators
        ):
          subgraph_operators += (
              tfl_flatbuffer_utils.get_subgraph_input_output_operators(subgraph)
          )
        total_operators += len(subgraph_operators)
        for op in subgraph_operators:
          subgraph_inds.extend(
              tfl_flatbuffer_utils.get_op_side_effect_subgraphs(op)
          )
      total_ops += value * total_operators
    return total_ops

  # TODO(b/330740605)- Collect multiple QSVs in one run to save compute.
  def calibrate(
      self,
      calibration_dataset: dict[str, Iterable[_SignatureInput]],
      model_recipe_manager: recipe_manager.RecipeManager,
      cache_output: bool = False,
  ) -> None:
    """Calibrates the model using the given dataset for a model signature.

    The process is
    0. Initialize quantization statistics values (QSVs) using the initialization
    function (from AlgorithmManager) for the op if needed.
    1. Invoke TFL interpreter on the calibration data.
    2. Go through each op, ask RecipeManager for the quantization setting
    of the op.
    3. Ask AlgorithmManager for the calibration function of the op given the
    quantization setting.
    4. Apply the function to the op to obtain quantization statistics (qsvs) for
    the tensors associated with the op.
    5. Update the global qsv dictionary
    6. Start another round of calibration.

    Args:
      calibration_dataset: A dictionary of input data for calibration for the
        given model signature.
      model_recipe_manager: A RecipeManager object that contains the
        quantization recipe.
      cache_output: Whether to cache the output of the model during the
        calibration process. This is useful if there are dependencies between
        signatures/models (e.g., decode requires encode output).
    """
    op_codes = self._flatbuffer_model.operatorCodes
    if self._model_qsvs:
      logging.warning(
          "Calibrator contains non-empty model qsvs, and the current"
          " calibration process will start on top of this state (i.e., update"
          " the existing qsvs). If this is an unintended behavior please call"
          " reset_model_qsvs to reset model qsvs."
      )

    total_ops = self._get_total_operations(calibration_dataset)
    with progress_utils.ProgressBar(
        total_steps=total_ops,
        description="Running Calibration:",
        disappear_on_finish=True,
        disable=total_ops
        < 1000,  # We skip the progress bar for small models and small datasets.
    ) as pbar:
      # TODO: b/329322226 - Enable parallel calibration.
      for signature_key, dataset in calibration_dataset.items():
        # Step0: get subgraph index.
        subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
            self._tfl_interpreter, signature_key
        )

        for data in dataset:
          # Initialize tensor names updated in this round of calibration.
          updated_tensor_names = set()

          # Step1: run tfl interpreter on subgraph to get tensor content.
          signature_output = tfl_interpreter_utils.invoke_interpreter_signature(
              self._tfl_interpreter, data, signature_key
          )
          if cache_output:
            self._cached_output.append(signature_output)

          # Step2: go through each op in subgraph to update quantization
          # statistic values.
          subgraphs_inds = [subgraph_idx]
          while subgraphs_inds:
            subgraph_ind = subgraphs_inds.pop()
            self._tensor_content_map.update(
                tfl_interpreter_utils.get_tensor_name_to_content_map(
                    self._tfl_interpreter, subgraph_ind
                )
            )
            subgraph = self._flatbuffer_model.subgraphs[subgraph_ind]
            graph_info = qtyping.GraphInfo(
                subgraph.tensors, self._flatbuffer_model.buffers
            )
            # Add input/output operators if they are not in the subgraph.
            if not any(
                isinstance(op, qtyping.IOOperator) for op in subgraph.operators
            ):
              subgraph.operators += (
                  tfl_flatbuffer_utils.get_subgraph_input_output_operators(
                      subgraph
                  )
              )
            for op in subgraph.operators:
              pbar.update_single_step()
              if isinstance(op, qtyping.IOOperator):
                op_key = op.op_key
              else:
                op_code = op_codes[op.opcodeIndex].builtinCode
                if op_code not in tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME:
                  continue
                op_key = tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME[op_code]
              # Step2.1: query the quantization_recipe to get op quantization
              # settings.
              op_scope = tfl_flatbuffer_utils.get_op_scope(op, subgraph.tensors)
              algorithm_name, _ = model_recipe_manager.get_quantization_configs(
                  op_key, op_scope
              )
              if algorithm_name == algorithm_manager.AlgorithmName.NO_QUANTIZE:
                continue
              if policy.is_non_quantizable_composite_op(op):
                continue

              # Step2.2: query algorithm_manager to get/call the related
              # calibration function.
              calibrate_func = algorithm_manager.get_quantization_func(
                  algorithm_name, op_key, qtyping.QuantizeMode.CALIBRATE
              )
              op_qsvs = calibrate_func(op, graph_info, self._tensor_content_map)
              # Step3: Update tensor qsvs with the new values. Ignore the tensor
              # names that are already updated in this round of calibration.
              op_updated_tensor_name = self._update_qsvs(
                  op_qsvs, updated_tensor_names
              )
              updated_tensor_names.update(op_updated_tensor_name)

              # Step4: Invoke any subgraphs invoked as a side effect of the op.
              subgraphs_inds.extend(
                  tfl_flatbuffer_utils.get_op_side_effect_subgraphs(op)
              )

        # Reset interpreter after one round of calibration.
        self._tfl_interpreter.reset_all_variables()

  def get_model_qsvs(self) -> dict[str, qtyping.QSV]:
    """Get the model qsvs.

    Returns:
      A dictionary of tensor name to QSV.
    """
    return self._model_qsvs

  def get_cached_output(self) -> list[_SignatureOutput]:
    """Get the cached output of the model."""
    return self._cached_output

  def clear_cached_output(self) -> None:
    """Clear the cached output of the model."""
    self._cached_output = []

  def reset_model_qsvs(self) -> None:
    """Reset the model qsvs."""
    self._model_qsvs = {}

  def load_model_qsvs(
      self, model_qsvs: Union[str, dict[str, qtyping.QSV]]
  ) -> None:
    """Load the model qsvs.

    Args:
      model_qsvs: A dictionary of tensor name to QSV or a path to a JSON file
        that contains the model qsvs (i.e., from save_calibration_result).
    """

    if isinstance(model_qsvs, str):
      self._model_qsvs = calibration_utils.load_calibration_results(model_qsvs)
    else:
      self._model_qsvs = copy.deepcopy(model_qsvs)

  def save_calibration_result(self, file_path: str) -> None:
    """Saves the calibration result to a json file."""
    with open(file_path, "w") as f:
      json.dump(self._model_qsvs, f, cls=calibration_utils.NumpyEncoder)

  def get_signature_list(self) -> list[str]:
    """Get the signature list of the model."""
    return self._tfl_interpreter.get_signature_list()

  def _update_qsvs(
      self,
      op_qsvs: dict[str, qtyping.QSV],
      ignore_tensor_names: set[str],
  ) -> set[str]:
    """Update the model qsvs with the new values.

    Args:
      op_qsvs: A dictionary of tensor name to QSV.
      ignore_tensor_names: A set of tensor names to ignore.

    Returns:
      A set of tensor names that are updated.
    """
    updated_tensor_names = set()
    for tensor_name, qsv in op_qsvs.items():
      if tensor_name in ignore_tensor_names:
        continue
      if tensor_name not in self._model_qsvs:
        self._model_qsvs[tensor_name] = qsv
      else:
        updated_qsv = self._qsv_update_func(self._model_qsvs[tensor_name], qsv)
        self._model_qsvs[tensor_name] = updated_qsv
      updated_tensor_names.add(tensor_name)
    return updated_tensor_names



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse
import pathlib
import tempfile

from absl.testing import absltest

from ai_edge_quantizer import aeq
from ai_edge_quantizer.utils import test_utils


class AeqTest(absltest.TestCase):

  def test_quantize(self):
    model_file = test_utils.get_path_to_datafile(
        "tests/models/conv_fc_mnist.tflite"
    )
    recipe_file = test_utils.get_path_to_datafile(
        "recipes/default_af32w8float_recipe.json"
    )

    with tempfile.TemporaryDirectory() as output_dir:
      aeq.main(
          argparse.Namespace(
              model_file=model_file,
              recipe_file=recipe_file,
              output_dir=output_dir,
              overwrite_outputs=False,
          )
      )

      model_basename = pathlib.Path(model_file).stem
      recipe_basename = pathlib.Path(recipe_file).stem
      output_filename = f"{model_basename}_{recipe_basename}.tflite"
      output_path = str(pathlib.Path(output_dir) / output_filename)

      self.assertTrue(pathlib.Path(output_path).exists())

      original_size = pathlib.Path(model_file).stat().st_size
      quantized_size = pathlib.Path(output_path).stat().st_size

      # Check whether the file size of the quantized model is roughly 4X smaller
      # than that of the float model
      self.assertLess(quantized_size, original_size * 0.35)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from absl.testing import parameterized
import numpy as np

import absl.testing.absltest as absltest
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import calibration_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_RNG = np.random.default_rng(66)

_CALIBRATION_DATASET = {
    "signature_1": [{
        "cache_0": np.zeros(shape=(1, 100, 4, 4), dtype=np.float32),
        "cache_1": np.zeros(shape=(1, 100, 4, 4), dtype=np.float32),
        "positions": np.zeros(shape=(1, 100), dtype=np.int32),
        "tokens": np.zeros(shape=(1, 100), dtype=np.int32),
    }],
    "signature_2": [{
        "cache_0": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
        "cache_1": _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
        "positions": (
            _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
        ),
        "tokens": _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32),
    }],
}


def _get_quant_parameters(
    quantized_model: bytes, signature_data: dict[str, list[str]]
) -> list[np.ndarray]:
  """Returns the quantization parameters from the quantized model."""
  quant_params = []
  tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
      quantized_model
  )
  for signature_key, signature_names in signature_data.items():
    signature_runner = tfl_interpreter.get_signature_runner(signature_key)

    for signature_name in signature_names:
      input_details = signature_runner.get_input_details()
      output_details = signature_runner.get_output_details()
      if signature_name in input_details.keys():
        quant_param = input_details[signature_name]["quantization_parameters"][
            "scales"
        ].squeeze()
        quant_params.append(quant_param)
      elif signature_name in output_details.keys():
        output_details = signature_runner.get_output_details()
        quant_param = output_details[signature_name]["quantization_parameters"][
            "scales"
        ].squeeze()
        quant_params.append(quant_param)
      else:
        raise ValueError(
            f"Signature name {signature_name} not found in the model."
        )
  return quant_params


class CalibrationQsvAlignmentUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="zero_smoothing_factor",
          smoothing_factor=0,
          expected_vals={"min": -1000, "max": 800},
      ),
      dict(
          testcase_name="one_smoothing_factor",
          smoothing_factor=1,
          expected_vals={"min": -10, "max": 8},
      ),
      dict(
          testcase_name="normal_smoothing_factor",
          smoothing_factor=0.99,
          expected_vals={"min": -19.9, "max": 15.92},
      ),
  )
  def test_update_tensor_qsv_moving_average(
      self, smoothing_factor, expected_vals
  ):
    old_qsv = {"min": -10, "max": 8}
    # Large values to mimic outlier.
    new_qsv = {"min": -1000, "max": 800}
    updated_qsv = calibration_utils.moving_average_update(
        old_qsv, new_qsv, smoothing_factor=smoothing_factor
    )
    self.assertAlmostEqual(updated_qsv["min"], expected_vals["min"])
    self.assertAlmostEqual(updated_qsv["max"], expected_vals["max"])

  @parameterized.named_parameters(
      dict(
          testcase_name="scalar",
          old_qsv={"min": -10, "max": 8},
          new_qsv={"min": -1000, "max": 1},
          expected_qsv={"min": -1000, "max": 8},
      ),
      dict(
          testcase_name="2darray",
          old_qsv={"min": [[-19], [20]], "max": [[21], [250]]},
          new_qsv={"min": [[-1000], [25]], "max": [[33], [100]]},
          expected_qsv={"min": [[-1000], [20]], "max": [[33], [250]]},
      ),
  )
  def test_update_tensor_qsv_min_max(self, old_qsv, new_qsv, expected_qsv):
    updated_qsv = calibration_utils.min_max_update(old_qsv, new_qsv)
    if isinstance(expected_qsv["min"], list):
      self.assertEqual(list(updated_qsv["min"]), expected_qsv["min"])
      self.assertEqual(list(updated_qsv["max"]), expected_qsv["max"])
    else:
      self.assertEqual(updated_qsv["min"], expected_qsv["min"])
      self.assertEqual(updated_qsv["max"], expected_qsv["max"])

  def test_load_calibration_results(self):
    temp_file = self.create_tempfile()
    temp_file.write_text('{"tensor1": {"min": [-1.0], "max": [1.0]}}')
    results = calibration_utils.load_calibration_results(temp_file.full_path)
    self.assertIn("tensor1", results)
    self.assertTrue(np.array_equal(results["tensor1"]["min"], [-1.0]))
    self.assertTrue(np.array_equal(results["tensor1"]["max"], [1.0]))

  def test_calibration_utils_init_fails(self):
    model_path = "non_existent_model.tflite"
    with self.assertRaisesWithPredicateMatch(
        Exception, lambda err: f"{model_path}" in str(err)
    ):
      calibration_utils.CalibrationQsvAlignmentUtils(model_path)

  def test_calibration_utils_init_succeeds(self):
    model_path = test_utils.get_path_to_datafile(
        "../tests/models/single_add.tflite"
    )
    calib_utils = calibration_utils.CalibrationQsvAlignmentUtils(model_path)
    self.assertNotEmpty(calib_utils._signature_runners)
    self.assertNotEmpty(calib_utils._same_as_input_scale_ops)

  def test_search_tensor_by_signature_name_succeeds_on_unconstrained_op(self):
    model_path = test_utils.get_path_to_datafile(
        "../tests/models/single_add.tflite"
    )
    expected_tensor_name = "PartitionedCall:0"
    calib_utils = calibration_utils.CalibrationQsvAlignmentUtils(model_path)
    tensor_name = calib_utils._search_tensor_by_signature_name(
        "serving_default", "add"
    )
    self.assertEqual(tensor_name, [expected_tensor_name])

  def test_search_tensor_by_signature_name_succeeds_on_constrained_op(self):
    model_path = test_utils.get_path_to_datafile(
        "../tests/models/single_slice.tflite"
    )
    expected_tensor_name = "slice_input_tensor:0"
    calib_utils = calibration_utils.CalibrationQsvAlignmentUtils(model_path)
    tensor_name = calib_utils._search_tensor_by_signature_name(
        "slice", "output_0"
    )
    self.assertEqual(tensor_name, [expected_tensor_name])

  def test_align_quant_stats_succeeds(self):
    model_path = test_utils.get_path_to_datafile(
        "../tests/models/toy_model_with_kv_cache_multi_signature.tflite"
    )
    recipe_path = test_utils.get_path_to_datafile(
        "../recipes/default_a8w8_recipe.json"
    )
    signature_data = {
        "signature_1": ["output_1_1"],
        "signature_2": ["output_1_1"],
    }

    # Obtain the calibration results.
    qt = quantizer.Quantizer(model_path, recipe_path)
    qsv = qt.calibrate(_CALIBRATION_DATASET)

    # First quantize the model without aligning the quantization parameters.
    quantized_model = qt.quantize(qsv).quantized_model
    quant_params = _get_quant_parameters(quantized_model, signature_data)
    self.assertFalse(
        all(x == quant_params[0] for x in quant_params)
    )  # not equal quantization params.

    # Align the quantization parameters and quantize again.
    calib_utils = calibration_utils.CalibrationQsvAlignmentUtils(model_path)
    calib_utils.align_quant_stats(qsv, signature_data)
    quantized_model = qt.quantize(qsv).quantized_model
    quant_params = _get_quant_parameters(quantized_model, signature_data)
    self.assertTrue(
        all(x == quant_params[0] for x in quant_params)
    )  # equal quantization params.

  def test_update_quant_stats_succeeds(self):
    model_path = test_utils.get_path_to_datafile(
        "../tests/models/toy_model_with_kv_cache_multi_signature.tflite"
    )
    recipe_path = test_utils.get_path_to_datafile(
        "../recipes/default_a8w8_recipe.json"
    )
    signature_data = {
        "signature_1": ["output_1_1"],
        "signature_2": ["output_1_1"],
    }

    # Obtain the calibration results.
    qt = quantizer.Quantizer(model_path, recipe_path)
    qsv = qt.calibrate(_CALIBRATION_DATASET)

    # First quantize the model without updating the `signature_1`.
    quantized_model = qt.quantize(qsv).quantized_model
    quant_params = _get_quant_parameters(quantized_model, signature_data)
    self.assertFalse(
        all(x == quant_params[0] for x in quant_params)
    )  # not equal quantization params.

    # Update the `signature_1` with stats from `signature_2`.
    calib_utils = calibration_utils.CalibrationQsvAlignmentUtils(model_path)
    min_val, max_val = calib_utils.align_quant_stats(  # for min and max only.
        qsv,
        {
            "signature_2": ["output_1_1"],
        },
    )
    calib_utils.update_quant_stats(
        qsv,
        {
            "signature_1": ["output_1_1"],
        },
        min_val,
        max_val,
    )
    quantized_model = qt.quantize(qsv).quantized_model
    quant_params = _get_quant_parameters(quantized_model, signature_data)
    self.assertTrue(
        all(x == quant_params[0] for x in quant_params)
    )  # equal quantization params.


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import dequantized_weight_recovery

_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig


class DequantizedWeightRecoveryTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._dummy_quantized_weights = np.array([
        [1, -2, 3, 4],
        [6, 7, -6, 5],
        [2, -6, -7, -4],
    ])
    self._dummy_op_info = qtyping.OpInfo(
        op=None,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=0,
        op_quant_config=qtyping.OpQuantizationConfig(),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="per-tensor-recovery",
          quantized_dimension=None,
          scale=np.array([0.1875]).reshape(1, 1),
      ),
      dict(
          testcase_name="channel0-recovery",
          quantized_dimension=0,
          scale=np.array([0.1875, 1e-4, 12.3]).reshape(3, 1),
      ),
      dict(
          testcase_name="channel1-recovery",
          quantized_dimension=1,
          scale=np.array([0.003, 1.234, 12.65, 2.24e-4]).reshape(1, 4),
      ),
  )
  def test_tensor_zp_scale_from_2d_dequantized_symmetric_weights_success(
      self, quantized_dimension, scale
  ):
    dequant_vals = scale * self._dummy_quantized_weights
    zp, recovered_scale = (
        dequantized_weight_recovery.get_zp_scale_from_dequantized_symmetric_weights(
            dequant_vals, quantized_dimension
        )
    )
    self.assertEqual(recovered_scale.shape, scale.shape)
    self.assertSequenceAlmostEqual(recovered_scale.flatten(), scale.flatten())
    # Zero point should be zero for symmetric quantization.
    self.assertEqual(np.sum(zp), 0)
    self.assertEqual(zp.shape, scale.shape)

  @parameterized.named_parameters(
      dict(
          testcase_name="per-tensor-recovery",
          quantized_dimension=None,
          scale=np.array([0.1875]).reshape(1, 1),
      ),
      dict(
          testcase_name="channel0-recovery",
          quantized_dimension=0,
          scale=np.array([0.1875, 1e-4, 12.3]).reshape(3, 1, 1),
      ),
      dict(
          testcase_name="channel1-recovery",
          quantized_dimension=1,
          scale=np.array([0.003, 1.234]).reshape(1, 2, 1),
      ),
  )
  def test_tensor_zp_scale_from_3d_dequantized_symmetric_weights_success(
      self, quantized_dimension, scale
  ):
    dequant_vals = scale * self._dummy_quantized_weights.reshape(3, 2, 2)
    zp, recovered_scale = (
        dequantized_weight_recovery.get_zp_scale_from_dequantized_symmetric_weights(
            dequant_vals, quantized_dimension
        )
    )
    with self.subTest("shapes_match"):
      self.assertEqual(recovered_scale.shape, scale.shape)
      self.assertEqual(zp.shape, scale.shape)
    with self.subTest("scale_value_match"):
      self.assertSequenceAlmostEqual(recovered_scale.flatten(), scale.flatten())
    with self.subTest("zp_is_zero"):
      # Zero point should be zero for symmetric quantization.
      self.assertEqual(np.sum(zp), 0)

  @parameterized.named_parameters(
      dict(testcase_name="negative_dimension", quantized_dimension=-1),
      dict(testcase_name="too_large_dimension", quantized_dimension=2),
  )
  def test_tensor_zp_scale_from_2d_dequantized_symmetric_weights_raises_error_for_invalid_quantized_dimension(
      self, quantized_dimension
  ):
    dequant_vals = self._dummy_quantized_weights * 1.02
    with self.assertRaisesRegex(
        ValueError, "quantized_dimension must be 0, 1, or None. Got"
    ):
      dequantized_weight_recovery.get_zp_scale_from_dequantized_symmetric_weights(
          dequant_vals, quantized_dimension
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="tensor-recovery-tensor-quant",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              granularity=qtyping.QuantGranularity.TENSORWISE,
          ),
          scale=np.array([0.1875]).reshape(1, 1),
      ),
      dict(
          testcase_name="channel-recovery-channel-quant",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              granularity=qtyping.QuantGranularity.CHANNELWISE,
          ),
          scale=np.array([0.1875, 1e-4, 12.3]).reshape(3, 1),
      ),
      dict(
          testcase_name="channel-recovery-excessive-bits",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=8,  # int4 is enough for the sample weights.
              granularity=qtyping.QuantGranularity.CHANNELWISE,
          ),
          scale=np.array([0.1875, 1e-4, 12.3]).reshape(3, 1),
      ),
  )
  def test_get_tensor_quant_params_success_with_dequantized_weights(
      self, tensor_quant_config, scale
  ):
    dequant_vals = scale * self._dummy_quantized_weights
    tensor_quant_params = dequantized_weight_recovery.get_tensor_quant_params(
        self._dummy_op_info, tensor_quant_config, dequant_vals
    )

    if tensor_quant_config.granularity is qtyping.QuantGranularity.TENSORWISE:
      self.assertIsNone(tensor_quant_params.quantized_dimension)
    else:
      self.assertEqual(tensor_quant_params.quantized_dimension, 0)

    recovered_scale = tensor_quant_params.scale
    self.assertEqual(recovered_scale.shape, scale.shape)
    self.assertSequenceAlmostEqual(recovered_scale.flatten(), scale.flatten())

    # Zero point should be zero for symmetric quantization.
    recovered_zp = tensor_quant_params.zero_point
    self.assertEqual(np.sum(recovered_zp), 0)
    self.assertEqual(recovered_zp.shape, scale.shape)

  def test_get_tensor_quant_params_success_with_qsv(self):
    # Fall back to naive_min_max_quantize.py for non-weight tensors.
    tensor_quant_params = dequantized_weight_recovery.get_tensor_quant_params(
        self._dummy_op_info,
        tensor_quant_config=qtyping.TensorQuantizationConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        tensor_qsv={
            "min": np.array([-1]),
            "max": np.array([1]),
        },
    )

    self.assertIsNone(tensor_quant_params.quantized_dimension)
    recovered_scale = tensor_quant_params.scale
    self.assertEqual(recovered_scale.shape, (1,))
    self.assertSequenceAlmostEqual(recovered_scale.flatten(), [1 / 127])

    # Zero point should be zero for symmetric quantization.
    recovered_zp = tensor_quant_params.zero_point
    self.assertEqual(np.sum(recovered_zp), 0)
    self.assertEqual(recovered_zp.shape, (1,))

  @parameterized.named_parameters(
      dict(
          testcase_name="recovery_on_wrong_dimension",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              granularity=qtyping.QuantGranularity.CHANNELWISE,
          ),
          scale=np.array([0.003, 1.234, 12.65, 2.24e-4]).reshape(1, 4),
      ),
      dict(
          testcase_name="tensor_recovery_for_channel_quantization",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              granularity=qtyping.QuantGranularity.TENSORWISE,
          ),
          scale=np.array([0.1875, 1e-2, 12.3]).reshape(3, 1),
      ),
      dict(
          testcase_name="insufficient_bits",
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=2,
              granularity=qtyping.QuantGranularity.CHANNELWISE,
          ),
          scale=np.array([0.1875, 1e-2, 12.3]).reshape(3, 1),
      ),
  )
  def test_get_tensor_quant_params_raises_error_big_recovery_error(
      self, tensor_quant_config, scale
  ):
    dequant_vals = scale * self._dummy_quantized_weights
    with self.assertRaisesRegex(
        RuntimeError,
        "Failed to recover the original quantized values from dequantized"
        " values. Max diff between recovered and original values: ",
    ):
      dequantized_weight_recovery.get_tensor_quant_params(
          self._dummy_op_info, tensor_quant_config, dequant_vals
      )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import duplicate_buffer
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('..')


class DuplicateBufferTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/weight_sharing_fcs.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)

  def _get_transformation_input(
      self, subgraph_idx: int, tensor_idx: int
  ) -> transformation_utils.TransformationInput:
    return transformation_utils.TransformationInput(
        tensor_id=tensor_idx,
        buffers=self.model.buffers,
        # Dummy params below.
        op_codes=self.model.operatorCodes,
        subgraph=self.model.subgraphs[subgraph_idx],
        producer=-1,
        consumers=[],
        quant_params=qtyping.UniformQuantParams(
            num_bits=8,
            quantized_dimension=None,
            scale=np.ones(1),
            zero_point=np.zeros(1),
        ),
    )

  def test_constant_buffer_is_correctly_duplicated(self):
    # Duplicate the FC weight tensor in the second subgraph.
    subgraph_idx = 1
    subgraph = self.model.subgraphs[subgraph_idx]
    weight_tensor_idx = 1
    prev_buffer_id = subgraph.tensors[weight_tensor_idx].buffer
    prev_num_buffers = len(self.model.buffers)
    transformation_input = self._get_transformation_input(
        subgraph_idx, weight_tensor_idx
    )
    transformation_info = duplicate_buffer.duplicate_buffer(
        transformation_input
    )
    self.assertEqual(transformation_info.op_id, 0)
    self.assertEqual(transformation_info.num_ops_added, 0)
    self.assertEqual(transformation_info.output_tensor_id, 1)
    # Check that a new buffer was added.
    self.assertLen(self.model.buffers, prev_num_buffers + 1)
    # Check that the new buffer is used by the weight tensor.
    new_buffer_id = len(self.model.buffers) - 1
    self.assertEqual(subgraph.tensors[weight_tensor_idx].buffer, new_buffer_id)
    # Check that the new buffer has the same data as the original one.
    self.assertTrue(
        np.all(
            np.frombuffer(
                self.model.buffers[new_buffer_id].data,
                dtype=np.float32,
            )
            == np.frombuffer(
                self.model.buffers[prev_buffer_id].data,
                dtype=np.float32,
            )
        )
    )

  def test_duplicate_buffer_raises_error_when_tensor_is_not_constant(self):
    # Duplicate the FC input tensor in the second subgraph.
    subgraph_idx = 1
    weight_tensor_idx = 0
    transformation_input = self._get_transformation_input(
        subgraph_idx, weight_tensor_idx
    )
    with self.assertRaisesRegex(
        ValueError,
        'Duplicate Buffer transformation supports only constant tensors.',
    ):
      duplicate_buffer.duplicate_buffer(transformation_input)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Util functions for TFL interpreter."""

from typing import Any, Optional, Union

import ml_dtypes
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_litert import interpreter as tfl  # pylint: disable=g-direct-tensorflow-import
import os

DEFAULT_SIGNATURE_KEY = "serving_default"

_Numeric = Union[int, float]


def create_tfl_interpreter(
    tflite_model: Union[str, bytes],
    allocate_tensors: bool = True,
    use_xnnpack: bool = True,
    num_threads: int = 16,
    preserve_all_tensors: bool = True,
) -> tfl.Interpreter:
  """Creates a TFLite interpreter from a model file.

  Args:
    tflite_model: Model file path or bytes.
    allocate_tensors: Whether to allocate tensors.
    use_xnnpack: Whether to use the XNNPACK delegate for the interpreter.
    num_threads: The number of threads to use for the interpreter.
    preserve_all_tensors: Whether to preserve all tensors. If False, only input
      and output tensors are preserved.

  Returns:
    A TFLite interpreter.
  """
  if isinstance(tflite_model, str):
    with open(tflite_model, "rb") as f:
      tflite_model = f.read()

  if use_xnnpack:
    op_resolver = tfl.OpResolverType.BUILTIN
  else:
    op_resolver = tfl.OpResolverType.BUILTIN_WITHOUT_DEFAULT_DELEGATES
  tflite_interpreter = tfl.Interpreter(
      model_content=bytes(tflite_model),
      num_threads=num_threads,
      experimental_op_resolver_type=op_resolver,
      experimental_preserve_all_tensors=preserve_all_tensors,
  )
  if allocate_tensors:
    tflite_interpreter.allocate_tensors()
  return tflite_interpreter


def has_quantization_params(tensor_detail: dict[str, Any]) -> bool:
  """Checks if a tensor contains quantization parameters.

  Args:
    tensor_detail: A dictionary of tensor details.

  Returns:
    True if the tensor is quantized.
  """
  quant_params = tensor_detail["quantization_parameters"]
  return bool(len(quant_params["scales"]))


def invoke_interpreter_signature(
    tflite_interpreter: tfl.Interpreter,
    signature_input_data: dict[str, Any],
    signature_key: Optional[str] = None,
    quantize_input: bool = True,
) -> dict[str, np.ndarray]:
  """Invokes the TFLite interpreter through signature runner.

  Args:
    tflite_interpreter: A TFLite interpreter.
    signature_input_data: The input data for the signature.
    signature_key: The signature key.
    quantize_input: Whether to quantize the input data.

  Returns:
    The output data of the signature.
  """
  # Make a copy to avoid in-place modification.
  signature_input = signature_input_data.copy()
  signature_runner = tflite_interpreter.get_signature_runner(signature_key)
  for input_name, input_detail in signature_runner.get_input_details().items():
    input_data = signature_input[input_name]
    if (
        input_data.dtype == np.float32
        and has_quantization_params(input_detail)
        and quantize_input
    ):
      quant_params = qtyping.UniformQuantParams.from_tfl_tensor_details(
          input_detail
      )
      signature_input[input_name] = uniform_quantize_tensor.uniform_quantize(
          input_data, quant_params
      )
  return signature_runner(**signature_input)


def invoke_interpreter_once(
    tflite_interpreter: tfl.Interpreter,
    input_data_list: list[Any],
    quantize_input: bool = True,
):
  """Invokes the TFLite interpreter once.

  Args:
    tflite_interpreter: A TFLite interpreter.
    input_data_list: A list of input data.
    quantize_input: Whether to quantize the input data.
  """
  if len(input_data_list) != len(tflite_interpreter.get_input_details()):
    raise ValueError(
        "Input data must be a list with each element match the input sequence"
        " defined in .tflite. If the model has only one input, wrap it with a"
        " list (e.g., [input_data])"
    )
  for i, input_data in enumerate(input_data_list):
    input_details = tflite_interpreter.get_input_details()[i]
    if (
        input_data.dtype == np.float32
        and has_quantization_params(input_details)
        and quantize_input
    ):
      quant_params = qtyping.UniformQuantParams.from_tfl_tensor_details(
          input_details
      )
      input_data = uniform_quantize_tensor.uniform_quantize(
          input_data, quant_params
      )
    tflite_interpreter.set_tensor(input_details["index"], input_data)
  tflite_interpreter.invoke()


def get_tensor_data(
    tflite_interpreter: Any,
    tensor_detail: dict[str, Any],
    subgraph_index: int = 0,
    dequantize: bool = True,
) -> np.ndarray:
  """Gets the tensor data from a TFLite interpreter.

  Args:
    tflite_interpreter: A TFLite interpreter.
    tensor_detail: A dictionary of tensor details.
    subgraph_index: The index of the subgraph that the tensor belongs to.
    dequantize: Whether to dequantize the quantized tensor data.

  Returns:
    The tensor data.
  """
  tensor_data = tflite_interpreter.get_tensor(
      tensor_detail["index"], subgraph_index
  )
  if (
      tensor_data.dtype != np.float32
      and has_quantization_params(tensor_detail)
      and dequantize
  ):
    quant_params = qtyping.UniformQuantParams.from_tfl_tensor_details(
        tensor_detail
    )
    tensor_data = uniform_quantize_tensor.uniform_dequantize(
        tensor_data,
        quant_params,
    )
  return tensor_data


def get_tensor_name_to_content_map(
    tflite_interpreter: Any, subgraph_index: int = 0, dequantize: bool = False
) -> dict[str, Any]:
  """Gets internal tensors from a TFLite interpreter for a given subgraph.

  Note the data will be copied to the returned dictionary, increasing the
  memory usage.

  Args:
    tflite_interpreter: A TFLite interpreter.
    subgraph_index: The index of the subgraph that the tensor belongs to.
    dequantize: Whether to dequantize the tensor data.

  Returns:
    A dictionary of internal tensors.
  """
  tensors = {}
  for tensor_detail in tflite_interpreter.get_tensor_details(subgraph_index):
    # Don't return temporary, unnamed tensors.
    if not tensor_detail["name"]:
      continue

    # Don't return tensors where any dimension of the shape is 0.
    if not np.all(tensor_detail["shape"]):
      continue

    tensors[tensor_detail["name"]] = get_tensor_data(
        tflite_interpreter, tensor_detail, subgraph_index, dequantize
    )
  return tensors


def get_tensor_name_to_details_map(
    tflite_interpreter: Any, subgraph_index: int = 0
) -> dict[str, Any]:
  """Gets internal tensors from a TFLite interpreter for a given subgraph.

  Args:
    tflite_interpreter: A TFLite interpreter.
    subgraph_index: The index of the subgraph that the tensor belongs to.

  Returns:
    A dictionary of internal tensors.
  """
  tensor_name_to_detail = {}
  for tensor_detail in tflite_interpreter.get_tensor_details(subgraph_index):
    # Don't return temporary, unnamed tensors or scratch tensors.
    # tensor_detail doesn't include the allocation size (bytes) or an
    # indicator of scratch tensors, so use the name to filter them out.
    if not tensor_detail["name"] or "scratch" in tensor_detail["name"]:
      continue
    tensor_name_to_detail[tensor_detail["name"]] = tensor_detail
  return tensor_name_to_detail


def get_input_tensor_names(
    tflite_model: Union[str, bytes], signature_name: Optional[str] = None
) -> list[str]:
  """Gets input tensor names from a TFLite model for a signature.

  Args:
    tflite_model: Model file path or bytes.
    signature_name: The signature name that the input tensors belong to.

  Returns:
    A list of input tensor names.
  """

  tfl_interpreter = create_tfl_interpreter(tflite_model, allocate_tensors=False)
  signature_runner = tfl_interpreter.get_signature_runner(signature_name)
  input_tensor_names = []
  for _, input_detail in signature_runner.get_input_details().items():
    input_tensor_names.append(input_detail["name"])
  return input_tensor_names


def get_output_tensor_names(
    tflite_model: Union[str, bytes], signature_name: Optional[str] = None
) -> list[str]:
  """Gets output tensor names from a TFLite model for a signature.

  Args:
    tflite_model: Model file path or bytes.
    signature_name: The signature name that the output tensors belong to.

  Returns:
    A list of output tensor names.
  """
  tfl_interpreter = create_tfl_interpreter(tflite_model, allocate_tensors=False)
  signature_runner = tfl_interpreter.get_signature_runner(signature_name)
  output_tensor_names = []
  for _, output_detail in signature_runner.get_output_details().items():
    output_tensor_names.append(output_detail["name"])
  return output_tensor_names


def get_constant_tensor_names(
    tflite_model: Union[str, bytes],
    subgraph_index: int = 0,
    min_constant_size: int = 1,
) -> list[str]:
  """Gets constant tensor names from a TFLite model for a subgraph.

  Note that this function acts on subgraph level, not signature level. This is
  because it is non-trivial to track constant tensors for a signature without
  running it.

  Args:
    tflite_model: Model file path or bytes.
    subgraph_index: The index of the subgraph that the tensor belongs to.
    min_constant_size: The minimum size of a constant tensor.

  Returns:
    A list of names for constant tensor that bigger than min_constant_size and a
    list of names for constant tensor that smaller than min_constant_size.
  """
  tfl_interpreter = create_tfl_interpreter(tflite_model, allocate_tensors=False)
  const_tensor_names = []
  for tensor_detail in tfl_interpreter.get_tensor_details(subgraph_index):
    if tensor_detail["dtype"] == np.object_:
      continue
    try:
      tensor_data = get_tensor_data(
          tfl_interpreter, tensor_detail, subgraph_index
      )
      if tensor_data.size >= min_constant_size:
        const_tensor_names.append(tensor_detail["name"])
    except ValueError:
      continue
  return const_tensor_names


def get_signature_main_subgraph_index(
    tflite_interpreter: tfl.Interpreter,
    signature_key: Optional[str] = None,
) -> int:
  """Gets the main subgraph index of a signature.

  Args:
    tflite_interpreter: A TFLite interpreter.
    signature_key: The signature key.

  Returns:
    The main subgraph index of the signature.
  """
  signature_runner = tflite_interpreter.get_signature_runner(signature_key)
  return signature_runner._subgraph_index  # pylint:disable=protected-access


def _create_random_normal(
    rng: np.random.Generator,
    shape: tuple[int, ...],
    dtype: np.dtype,
) -> np.ndarray:
  """Creates a random normal dataset sample for given input details."""
  return rng.normal(size=shape).astype(dtype)


def _create_random_uniform(
    rng: np.random.Generator,
    shape: tuple[int, ...],
    dtype: np.dtype,
    min_value: float = 0.0,
    max_value: float = 1.0,
) -> np.ndarray:
  """Creates a random uniform dataset sample for given input details."""
  return rng.uniform(min_value, max_value, size=shape).astype(dtype)


def _create_random_integers(
    rng: np.random.Generator,
    shape: tuple[int, ...],
    dtype: np.dtype,
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
) -> np.ndarray:
  """Creates a random integer dataset sample for given input details."""
  if min_value is None:
    min_value = np.iinfo(dtype).min
  endpoint = False
  if max_value is None:
    max_value = np.iinfo(dtype).max
    endpoint = True
  return rng.integers(
      min_value, max_value, size=shape, dtype=dtype, endpoint=endpoint
  )


def _create_random_bool(
    rng: np.random.Generator,
    shape: tuple[int, ...],
    dtype: np.dtype,
) -> np.ndarray:
  """Creates a random bool dataset sample for given input details."""
  return rng.choice([True, False], size=shape, replace=True).astype(dtype)


def create_random_dataset(
    input_details: dict[str, Any],
    num_samples: int,
    random_seed: Union[int, np._typing.ArrayLike],
    min_max_range: Optional[tuple[_Numeric, _Numeric]] = None,
) -> list[dict[str, Any]]:
  """Creates a random normal dataset for given input details.

  Args:
    input_details: A dictionary of input details.
    num_samples: The number of samples to generate.
    random_seed: The random seed to use.
    min_max_range: The min and max of the input range.

  Returns:
    A list of dictionaries, each containing a sample of input data (for all
    signatures).
  """
  rng = np.random.default_rng(random_seed)
  dataset = []
  for _ in range(num_samples):
    input_data = {}
    for arg_name, input_tensor in input_details.items():
      dtype = input_tensor["dtype"]
      shape = input_tensor["shape"]
      if np.issubdtype(dtype, np.integer):
        if min_max_range is None:
          new_data = _create_random_integers(rng, shape, dtype)
        else:
          min_value, max_value = min_max_range
          new_data = _create_random_integers(
              rng, shape, dtype, min_value, max_value
          )
      elif dtype in (np.float32, ml_dtypes.bfloat16):
        if min_max_range is None:
          new_data = _create_random_normal(rng, shape, dtype)
        else:
          min_value, max_value = min_max_range
          new_data = _create_random_uniform(
              rng, shape, dtype, min_value, max_value
          )
      elif dtype == np.bool:
        new_data = _create_random_bool(rng, shape, dtype)
      else:
        raise ValueError(f"Unsupported dtype: {input_tensor['dtype']}")
      input_data[arg_name] = new_data
    dataset.append(input_data)
  return dataset


def create_random_normal_input_data(
    tflite_model: Union[str, bytes],
    num_samples: int = 4,
    random_seed: int = 666,
    min_max_range: Optional[tuple[_Numeric, _Numeric]] = None,
) -> dict[str, list[dict[str, Any]]]:
  """Creates a random normal dataset for a signature runner.

  Args:
    tflite_model: TFLite model path or bytearray.
    num_samples: Number of input samples to be generated.
    random_seed: Random seed to be used for function.
    min_max_range: The min and max of the input range.

  Returns:
    A list of inputs to the given interpreter, for a single interpreter we may
    have multiple signatures so each set of inputs is also represented as
    list.
  """
  tfl_interpreter = create_tfl_interpreter(tflite_model)
  signature_defs = tfl_interpreter.get_signature_list()
  signature_keys = list(signature_defs.keys())
  test_data = {}
  for signature_key in signature_keys:
    signature_runner = tfl_interpreter.get_signature_runner(signature_key)
    input_details = signature_runner.get_input_details()
    test_data[signature_key] = create_random_dataset(
        input_details,
        num_samples,
        random_seed,
        min_max_range,
    )
  return test_data



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for transformation_performer."""

import copy
import pathlib

import numpy as np

from absl.testing import parameterized
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import transformation_performer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_QTransf = qtyping.QuantTransformation


TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(".")


class TransformationPerformerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._transformation_performer = (
        transformation_performer.TransformationPerformer()
    )
    self._test_model = tfl_flatbuffer_utils.read_model(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/conv_fc_mnist.tflite"
        )
    )

  def test_apply_single_insert_dequant(self):
    """test for _apply_transformation."""
    self._transformation_performer._create_op_id_map(self._test_model)
    instructions = qtyping.TensorTransformationInsts(
        tensor_name="sequential/conv2d/Relu;sequential/conv2d/BiasAdd;"
        + "sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp1",
        subgraph_id=0,
        instructions=[
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                tensor_id=7,
                producer=0,
                consumers=[1],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
            qtyping.TransformationInst(
                transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                tensor_id=7,
                producer=0,
                consumers=[1],
                parameters=qtyping.UniformQuantParams(
                    8, None, np.array([1]), np.array([0])
                ),
            ),
        ],
    )
    self._transformation_performer._apply_single_transformation(
        instructions, 0, self._test_model
    )
    subgraph = self._test_model.subgraphs[0]
    self.assertIn(b"_dequant", subgraph.tensors[13].name)
    self.assertEqual(
        subgraph.operators[1].opcodeIndex,
        len(self._test_model.operatorCodes) - 1,
    )
    self.assertEqual(subgraph.operators[2].inputs[0], 13)

  def test_create_op_id_map(self):
    """test for _create_op_id_map."""
    self._transformation_performer._create_op_id_map(self._test_model)
    op_id_map = self._transformation_performer._original_op_id_map
    self.assertLen(op_id_map, 1)
    self.assertLen(op_id_map[0], 6)
    for index, op_id in enumerate(op_id_map[0]):
      self.assertEqual(op_id, index)

  def test_update_op_id_map_changing_value(self):
    """test for _update_op_id_map."""
    self._transformation_performer._create_op_id_map(self._test_model)
    self._transformation_performer._update_op_id_map(0, 1, 6)
    op_id_map = self._transformation_performer._original_op_id_map
    self.assertLen(op_id_map, 1)
    self.assertLen(op_id_map[0], 6)
    for index in range(1, len(op_id_map[0])):
      self.assertEqual(op_id_map[0][index], index + 6)

  def test_update_op_id_map_not_changing_value(self):
    """test for _update_op_id_map."""
    self._transformation_performer._create_op_id_map(self._test_model)
    self._transformation_performer._update_op_id_map(0, 0, 0)
    op_id_map = self._transformation_performer._original_op_id_map
    self.assertLen(op_id_map, 1)
    self.assertLen(op_id_map[0], 6)
    for index, op_id in enumerate(op_id_map[0]):
      self.assertEqual(op_id, index)

  def test_update_op_id_map_not_changing_value_single_op_model(self):
    """test for _update_op_id_map."""
    model = tfl_flatbuffer_utils.read_model(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/single_fc_bias.tflite"
        )
    )
    self._transformation_performer._create_op_id_map(model)
    instruction = qtyping.TransformationInst(
        transformation=qtyping.QuantTransformation.QUANTIZE_TENSOR,
        tensor_id=0,
        producer=0,
        consumers=[-1],
        parameters=qtyping.UniformQuantParams(
            8, None, np.array([1]), np.array([0])
        ),
    )
    producer = self._transformation_performer._get_updated_producer_id(
        instruction.producer, 0
    )
    consumers = self._transformation_performer._get_updated_consumer_ids(
        instruction.consumers, 0
    )
    self.assertEqual(producer, 0)
    self.assertEqual(consumers, [-1])

  @parameterized.named_parameters(
      dict(
          testcase_name="test_no_update",
          prev_trans_idx=0,
          instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          trans_info=qtyping.TransformationInfo(0, 0, 0),
          expected_instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected_added_op_id_map=[[]],
      ),
      dict(
          testcase_name="test_no_matching_consumer",
          prev_trans_idx=0,
          instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          trans_info=qtyping.TransformationInfo(2, 2, 13),
          expected_instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[2],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected_added_op_id_map=[[3]],
      ),
      dict(
          testcase_name="test_insert_one_op",
          prev_trans_idx=0,
          instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          trans_info=qtyping.TransformationInfo(1, 1, 13),
          expected_instructions=[
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                  tensor_id=0,
                  producer=0,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
              qtyping.TransformationInst(
                  transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                  tensor_id=13,
                  producer=6,
                  consumers=[1],
                  parameters=qtyping.UniformQuantParams(
                      8, None, np.array([1]), np.array([0])
                  ),
              ),
          ],
          expected_added_op_id_map=[[1]],
      ),
  )
  def test_update_instructions(
      self,
      prev_trans_idx,
      instructions,
      trans_info,
      expected_instructions,
      expected_added_op_id_map,
  ):
    """test for _update_instructions."""
    self._transformation_performer._create_op_id_map(self._test_model)
    self._transformation_performer._update_instructions(
        prev_trans_idx, instructions, 0, trans_info
    )
    self.assertSequenceEqual(instructions, expected_instructions)
    self.assertListEqual(
        self._transformation_performer._added_op_id_map,
        expected_added_op_id_map,
    )

  def test_update_instructions_updates_tensor_id_after_duplicate_tensor(self):
    def get_test_instruction(transformation, consumers):
      return qtyping.TransformationInst(
          transformation=transformation,
          consumers=consumers,
          # Dummy values below.
          tensor_id=0,
          producer=0,
          parameters=qtyping.UniformQuantParams(
              8, None, np.array([1]), np.array([0])
          ),
      )

    instructions = [
        get_test_instruction(_QTransf.DUPLICATE_TENSOR, consumers=[1]),
        get_test_instruction(_QTransf.ADD_QUANTIZE, consumers=[1]),
        get_test_instruction(_QTransf.ADD_DEQUANTIZE, consumers=[1]),
        get_test_instruction(_QTransf.QUANTIZE_TENSOR, consumers=[2]),
    ]
    # Simulate a situation as if the first instruction (duplicate tensor) was
    # applied.
    subgraph_id = 0
    duplicated_tensor_id = 13
    prev_trans_idx = 0
    trans_info = qtyping.TransformationInfo(
        # Copy of what duplicate_tensor.py returns.
        op_id=0,
        num_ops_added=0,
        output_tensor_id=duplicated_tensor_id,
    )
    self._transformation_performer._create_op_id_map(self._test_model)
    self._transformation_performer._update_instructions(
        prev_trans_idx, instructions, subgraph_id, trans_info
    )
    # Expecting the ops with the same consumers as in the DUPLICATE_TENSOR
    # instruction to use the new tensor id.
    expected_instructions = copy.deepcopy(instructions)
    expected_instructions[1].tensor_id = duplicated_tensor_id
    expected_instructions[2].tensor_id = duplicated_tensor_id
    self.assertSequenceEqual(instructions, expected_instructions)
    # Expecting no change to the op id map.
    self.assertListEqual(
        self._transformation_performer._added_op_id_map,
        [[]],
    )

  def test_transform_graph(self):
    """test for transform_graph."""
    instructions = {
        "sequential/conv2d/Relu;sequential/conv2d/BiasAdd;"
        + "sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp1": qtyping.TensorTransformationInsts(
            tensor_name="sequential/conv2d/Relu;sequential/conv2d/BiasAdd;"
            + "sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp1",
            subgraph_id=0,
            # Conv2d: op_id=0, output_tensor_id=7.
            # This should add two sequential dequants after the conv2d.
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=7,
                    producer=0,
                    consumers=[1],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=7,
                    producer=0,
                    consumers=[1],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
            ],
        ),
        "sequential/average_pooling2d/AvgPool": qtyping.TensorTransformationInsts(
            tensor_name="sequential/average_pooling2d/AvgPool",
            subgraph_id=0,
            # Avg_pool: op_id=1, output_tensor_id=8.
            # This should add two sequential dequants after the avg_pool.
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=8,
                    producer=1,
                    consumers=[2],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=8,
                    producer=1,
                    consumers=[2],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
            ],
        ),
    }
    self._transformation_performer.transform_graph(
        instructions, self._test_model
    )
    self.assertLen(self._test_model.subgraphs, 1)
    self.assertLen(self._test_model.subgraphs[0].operators, 10)
    # The original model has 13 tensors, each dequant adds 1 tensor.
    self.assertLen(self._test_model.subgraphs[0].tensors, 17)
    # Check that the dequant opcode is added to the model.
    self.assertEqual(
        self._test_model.subgraphs[0].operators[1].opcodeIndex,
        len(self._test_model.operatorCodes) - 1,
    )
    # Conv2d, dequant, dequant, avgpool, dequant, dequant, etc.
    expected_builtin_op_order = [3, 6, 6, 1, 6, 6, 22, 9, 9, 25]
    for i, op in enumerate(self._test_model.subgraphs[0].operators):
      op_code = self._test_model.operatorCodes[op.opcodeIndex].builtinCode
      self.assertEqual(op_code, expected_builtin_op_order[i])
    # Check that the first dequant input is connected to the conv2d output.
    self.assertEqual(self._test_model.subgraphs[0].operators[1].inputs[0], 7)
    # Output is a new tensor just added.
    self.assertEqual(self._test_model.subgraphs[0].operators[1].outputs[0], 13)
    # Second dequant has new tensors.
    self.assertEqual(self._test_model.subgraphs[0].operators[2].inputs[0], 13)
    self.assertEqual(self._test_model.subgraphs[0].operators[2].outputs[0], 14)
    # Avgpool's input is second dequant's output.
    self.assertEqual(self._test_model.subgraphs[0].operators[3].inputs[0], 14)
    # Avgpool's output remains the same.
    self.assertEqual(self._test_model.subgraphs[0].operators[3].outputs[0], 8)
    # Third dequant's output is a new tensor.
    self.assertEqual(self._test_model.subgraphs[0].operators[4].outputs[0], 15)
    # Fourth dequant.
    self.assertEqual(self._test_model.subgraphs[0].operators[5].inputs[0], 15)
    self.assertEqual(self._test_model.subgraphs[0].operators[5].outputs[0], 16)

    # Avgpool (op_id=1) and reshape (op_id=2) are bumped by 2 due to the two
    # dequants added after it.
    expected_op_id_map = [0, 3, 6, 7, 8, 9]
    self.assertEqual(
        self._transformation_performer._original_op_id_map[0],
        expected_op_id_map,
    )
    # New dequants are added at these indices.
    expected_added_op_id_map = [1, 2, 4, 5]
    self.assertEqual(
        self._transformation_performer._added_op_id_map[0],
        expected_added_op_id_map,
    )

  def test_op_insertion_at_input_and_output(self):
    """test for _update_op_id_map."""
    model = tfl_flatbuffer_utils.read_model(
        str(
            pathlib.Path(TEST_DATA_PREFIX_PATH)
            / "tests/models/single_fc_bias.tflite"
        )
    )
    self._transformation_performer._create_op_id_map(model)
    instructions = {
        # Fully_connected: op_id=0, input_tensor_id=0, output_tensor_id=3.
        # Add a new quantize op to the input of the fully_connected.
        "serving_default_input_2:0": qtyping.TensorTransformationInsts(
            tensor_name="serving_default_input_2:0",
            subgraph_id=0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                    tensor_id=0,
                    producer=-1,
                    consumers=[0],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
            ],
        ),
        # Add a new dequantize op to the output of the fully_connected.
        "StatefulPartitionedCall:0": qtyping.TensorTransformationInsts(
            tensor_name="StatefulPartitionedCall:0",
            subgraph_id=0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=3,
                    producer=0,
                    consumers=[-1],
                    parameters=qtyping.UniformQuantParams(
                        8, None, np.array([1]), np.array([0])
                    ),
                ),
            ],
        ),
    }
    self._transformation_performer.transform_graph(instructions, model)

    # Original fc (op_id=0) should be bumped to op_id=1.
    self.assertEqual(
        self._transformation_performer._original_op_id_map[0],
        [1],
    )
    # New quantize added at op_id=0, dequantize added at op_id=1.
    expected_added_op_id_map = [0, 2]
    self.assertEqual(
        self._transformation_performer._added_op_id_map[0],
        expected_added_op_id_map,
    )
    # Quantize, fully_connected, dequantize.
    expected_builtin_op_order = [114, 9, 6]
    for i, op in enumerate(model.subgraphs[0].operators):
      op_code = model.operatorCodes[op.opcodeIndex].builtinCode
      self.assertEqual(op_code, expected_builtin_op_order[i])


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Type hinting support for AI Edge Quantizer."""

import collections
from collections.abc import MutableMapping
import copy
import dataclasses
import enum
from typing import Any, Callable, Optional, Union

import numpy as np
from typing_extensions import TypeAlias


QSV: TypeAlias = MutableMapping[str, Any]


class TFLOperationName(str, enum.Enum):
  """TF Lite operation names."""

  ALL_SUPPORTED = '*'
  INPUT = 'INPUT'
  OUTPUT = 'OUTPUT'
  FULLY_CONNECTED = 'FULLY_CONNECTED'
  BATCH_MATMUL = 'BATCH_MATMUL'
  DEPTHWISE_CONV_2D = 'DEPTHWISE_CONV_2D'
  CONV_2D = 'CONV_2D'
  CONV_2D_TRANSPOSE = 'CONV_2D_TRANSPOSE'
  AVERAGE_POOL_2D = 'AVERAGE_POOL_2D'
  RESHAPE = 'RESHAPE'
  CUSTOM_OP = 'CUSTOM_OP'
  EMBEDDING_LOOKUP = 'EMBEDDING_LOOKUP'
  SOFTMAX = 'SOFTMAX'
  TANH = 'TANH'
  TRANSPOSE = 'TRANSPOSE'
  GELU = 'GELU'
  ADD = 'ADD'
  SUB = 'SUB'
  MUL = 'MUL'
  MEAN = 'MEAN'
  RSQRT = 'RSQRT'
  CONCATENATION = 'CONCATENATION'
  STRIDED_SLICE = 'STRIDED_SLICE'
  SPLIT = 'SPLIT'
  LOGISTIC = 'LOGISTIC'
  SLICE = 'SLICE'
  SUM = 'SUM'
  SELECT = 'SELECT'
  SELECT_V2 = 'SELECT_V2'
  DYNAMIC_UPDATE_SLICE = 'DYNAMIC_UPDATE_SLICE'
  STABLEHLO_COMPOSITE = 'STABLEHLO_COMPOSITE'
  PAD = 'PAD'
  SQUARED_DIFFERENCE = 'SQUARED_DIFFERENCE'
  MAX_POOL_2D = 'MAX_POOL_2D'
  RESIZE_BILINEAR = 'RESIZE_BILINEAR'
  RESIZE_NEAREST_NEIGHBOR = 'RESIZE_NEAREST_NEIGHBOR'
  GATHER_ND = 'GATHER_ND'
  PACK = 'PACK'
  UNPACK = 'UNPACK'
  DIV = 'DIV'
  BROADCAST_TO = 'BROADCAST_TO'
  SQRT = 'SQRT'
  GATHER = 'GATHER'
  HARD_SWISH = 'HARD_SWISH'
  MAXIMUM = 'MAXIMUM'
  PADV2 = 'PADV2'
  REDUCE_MIN = 'REDUCE_MIN'
  EQUAL = 'EQUAL'
  NOT_EQUAL = 'NOT_EQUAL'
  MIRROR_PAD = 'MIRROR_PAD'
  SPACE_TO_DEPTH = 'SPACE_TO_DEPTH'
  RELU = 'RELU'


class QuantizeMode(enum.Enum):
  CALIBRATE = 2
  MATERIALIZE = 3


class OpExecutionMode(str, enum.Enum):
  """How to execute the op."""

  WEIGHT_ONLY = 'WEIGHT_ONLY'
  DRQ = 'DRQ'  # Dynamic range quantization.
  SRQ = 'SRQ'  # Static range quantization.


class ComputePrecision(str, enum.Enum):
  """The precision of the compute operation."""

  INTEGER = 'INTEGER'
  FLOAT = 'FLOAT'


class TensorDataType(str, enum.Enum):
  INT = 'INT'
  FLOAT = 'FLOAT'


class QuantGranularity(str, enum.Enum):
  TENSORWISE = 'TENSORWISE'
  CHANNELWISE = 'CHANNELWISE'
  # Blockwise quantization with various block sizes.
  BLOCKWISE_32 = 'BLOCKWISE_32'
  BLOCKWISE_64 = 'BLOCKWISE_64'
  BLOCKWISE_128 = 'BLOCKWISE_128'
  BLOCKWISE_256 = 'BLOCKWISE_256'


class QuantTransformation(enum.Enum):
  """Operations associated with quantization for a tensor."""

  # Do nothing: float_tensor -> float_tensor.
  NO_QUANTIZE = 0
  # Add a quantize op: float_tensor -> Quantize Op -> quantized_tensor.
  ADD_QUANTIZE = 1
  # Add a dequantize op: quantized_tensor -> Dequantize Op -> float_tensor.
  ADD_DEQUANTIZE = 2
  # Quantize the float tensor: float_tensor -> quantized_tensor.
  QUANTIZE_TENSOR = 3
  # (Deprecated) Create pattern for emulated subchannel quantization,
  # only support fully connected op.
  EMULATED_SUBCHANNEL = 4
  # Duplicate the buffer.
  DUPLICATE_BUFFER = 5
  # Duplicate the tensor.
  DUPLICATE_TENSOR = 6
  # Insert the aeq.hadamard_rotation op.
  INSERT_HADAMARD_ROTATION = 7
  # Insert decomposed Hadamard rotation ops. This expresses the Hadamard
  # rotation as matrix multiplication with Hadamard matrices.
  INSERT_DECOMPOSED_HADAMARD_ROTATION = 8


@dataclasses.dataclass(frozen=True)
class UniformQuantParams:
  """Parameters for uniform quantization.

  Attributes:
    num_bits: Number of bits to quantize to (e.g. 8 for int8).
    quantized_dimension: The dimension to quantize.
    scale: The scale of the quantization.
    zero_point: The zero point of the quantization.
    symmetric: Whether the quantization is symmetric (force zero_point to be 0).
    quantized_data: The quantized data.
    block_size: The block size for blockwise quantization, block_size=0 meaning
      no blockwise quantization.
    hadamard: The Hadamard rotation parameters, if set.
  """

  class HadamardRotationParams:
    """Parameters for the Hadamard rotation.

    Attributes:
      random_binary_vector: The random binary vector for the Hadamard rotation.
        TODO(b/415392354): Randomization is an experimental feature that's
        currently not implemented yet hence this is always 1. We will add
        support or remove in the future.
      hadamard_size: The size of the Hadamard matrix.
    """

    random_binary_vector: np.ndarray
    hadamard_size: int

    def __init__(self, random_binary_vector: np.ndarray, hadamard_size: int):
      self.random_binary_vector = random_binary_vector
      self.hadamard_size = hadamard_size

    def __eq__(self, other):
      if other.__class__ is not self.__class__:
        return NotImplemented
      return (
          np.array_equal(self.random_binary_vector, other.random_binary_vector)
          and self.hadamard_size == other.hadamard_size
      )

  num_bits: int
  quantized_dimension: Optional[int]
  scale: np.ndarray
  zero_point: np.ndarray
  symmetric: bool = True
  quantized_data: Optional[np.ndarray] = None
  block_size: int = 0
  hadamard: Optional[HadamardRotationParams] = None

  @classmethod
  def from_tfl_tensor_details(cls, tensor_detail) -> 'UniformQuantParams':
    """Creates UniformQuantParams from TFLite tensor details.

    Args:
      tensor_detail: The tensor details from TFLite.

    Returns:
      UniformQuantParams.
    """
    quant_params = tensor_detail['quantization_parameters']
    data_type = tensor_detail['dtype']
    if data_type == np.int8:
      num_bits = 8
    elif data_type == np.int16:
      num_bits = 16
    elif data_type == np.int32:
      num_bits = 32
    elif data_type == np.int64:
      num_bits = 64
    else:
      raise ValueError(f'Unsupported data type: {data_type}')
    symmetric = sum(abs(quant_params['zero_points'])) == 0
    return cls(
        quantized_dimension=quant_params['quantized_dimension'],
        num_bits=num_bits,
        scale=quant_params['scales'],
        zero_point=quant_params['zero_points'],
        symmetric=symmetric,
        block_size=quant_params['block_size'],
    )

  def __eq__(self, other):
    if other.__class__ is not self.__class__:
      return NotImplemented
    return (
        self.num_bits == other.num_bits
        and self.quantized_dimension == other.quantized_dimension
        and np.array_equal(self.scale, other.scale)
        and np.array_equal(self.zero_point, other.zero_point)
        and self.symmetric == other.symmetric
        and _compare_array_or_none(self.quantized_data, other.quantized_data)
        and self.block_size == other.block_size
        and self.hadamard == other.hadamard
    )


@dataclasses.dataclass(frozen=True)
class NonLinearQuantParams:
  """Parameters for nonlinear quantization.

  Currently only used for fp16 quantization.

  Attributes:
    num_bits: Number of bits to quantize to (e.g. 16 for fp16).
    quantized_data: The quantized data.
    data_type: The data type of the tensor.
  """

  num_bits: int
  quantized_data: Optional[np.ndarray]
  data_type: TensorDataType = TensorDataType.FLOAT

  def __eq__(self, other):
    if other.__class__ is not self.__class__:
      return NotImplemented
    return (
        self.num_bits == other.num_bits
        and self.data_type == other.data_type
        and _compare_array_or_none(self.quantized_data, other.quantized_data)
    )


@dataclasses.dataclass(frozen=True)
class OpToTensorParams:
  """Tensor params authored from an associated op.

  Attributes:
    subgraph_op_id: The position of the op in the subgraph.
    transformations: The transformations to be applied to the tensor.
    parameters: The quantization parameters for the tensor.
  """

  subgraph_op_id: int
  transformations: list[QuantTransformation]
  parameters: Union[None, UniformQuantParams, NonLinearQuantParams] = None


@dataclasses.dataclass
class TensorTransformationParams:
  """Transformation info for a tensor.

  Every tensor in .tflite has the following property:
   * Produced by one source op (producer), except constant tensor or model
   input.
   * Consumed by one or many destination ops (consumer), except model output.

  Because users configure quantization settings in Op level
  `OpQuantizationConfig`, each tensor will receive transformation parameters
   * from the source op
   * from the destination ops
  """

  tensor_name: str
  producer: Optional[OpToTensorParams] = None
  consumers: Optional[list[OpToTensorParams]] = None

  def __copy__(self):
    return TensorTransformationParams(
        tensor_name=self.tensor_name,
        producer=self.producer,
        consumers=self.consumers[:] if self.consumers is not None else None,
    )


@dataclasses.dataclass(frozen=True)
class TensorQuantizationConfig:
  """Quantization configuration for a tensor.

  Attributes:
    num_bits: Number of bits to quantize to (e.g. 8 for int8).
    symmetric: Whether to perform symmetric or asymmetric quantization. In the
      symmetric quantization mode, the zero point is always 0.
    granularity: Whether to perform per-tensor, per-channel or per-block
      quantization.
    dtype: The data type of the tensor.
    algorithm_key: The algorithm key to use for quantization.
  """

  num_bits: int
  symmetric: bool = True
  granularity: QuantGranularity = QuantGranularity.TENSORWISE
  dtype: TensorDataType = TensorDataType.INT

  def to_dict(self) -> dict[str, Any]:
    """Converts ActivationQuantizationConfig to dict."""
    return dataclasses.asdict(
        self,
        dict_factory=lambda x: {  # pylint: disable=g-long-lambda
            k: v
            for (k, v) in x
            # Skip None and empty dict values.
            if v is not None and not (isinstance(v, dict) and not v)
        },
    )

  @classmethod
  def from_dict(cls, params: dict[str, Any]) -> 'TensorQuantizationConfig':
    """Converts a given dict to TensorQuantizationConfig."""
    params_copy = copy.deepcopy(params)
    # Process block_size config from legacy recipe.
    params_copy = _process_block_size(params_copy)
    return cls(**params_copy)


def _process_block_size(params: dict[str, Any]) -> dict[str, Any]:
  """Processes block size in the params."""
  block_size = params.pop('block_size', 0)
  if block_size > 0:
    if block_size == 32:
      params['granularity'] = QuantGranularity.BLOCKWISE_32
    elif block_size == 64:
      params['granularity'] = QuantGranularity.BLOCKWISE_64
    elif block_size == 128:
      params['granularity'] = QuantGranularity.BLOCKWISE_128
    elif block_size == 256:
      params['granularity'] = QuantGranularity.BLOCKWISE_256
    else:
      raise ValueError(f'Unsupported block size: {block_size}')
  return params


@dataclasses.dataclass(frozen=True)
class OpQuantizationConfig:
  """Configuration class to control the quantization process behavior.

  Default to float activations and weights.

  Attributes:
    activation_tensor_config: The quantization configuration for activation
      tensors in the op (i.e., runtime tensors).
    weight_tensor_config: The quantization configuration for weight tensor in
      the op.
    compute_precision: The precision of the compute operation.
    explicit_dequantize: Whether to add explicit dequantize op if compute
      precision is FLOAT, but weight is quantized.
    skip_checks: Whether to skip op quantization config checks. For advanced
      users only. If set, the quantizer will ignore all op configuration checks
      and forcefully quantize this op according to the user instructions even if
      it's not supported in the TFLite runtime.
    min_weight_elements: The minimum number of elements in the weight tensor to
      be quantized.
  """

  activation_tensor_config: Optional[TensorQuantizationConfig] = None
  # Bias tensor quantization is deduced from activation/weight config.
  # e.g., int8A X int8W => int32 bias.
  weight_tensor_config: Optional[TensorQuantizationConfig] = None
  compute_precision: ComputePrecision = ComputePrecision.FLOAT
  # TODO: b/359647578 - Set default to True.
  explicit_dequantize: bool = False
  skip_checks: bool = False
  min_weight_elements: int = 0

  def __post_init__(self):
    if (
        self.activation_tensor_config is None
        or self.weight_tensor_config is None
    ):
      return
    # Make sure the setting is valid.
    if (
        self.activation_tensor_config.dtype == TensorDataType.INT
        and self.weight_tensor_config.dtype == TensorDataType.FLOAT
    ):
      raise ValueError(
          'An op can not be set to have integer activation but float weights!'
      )
    if (
        # SRQ compliance check for the config.
        self.activation_tensor_config.dtype == TensorDataType.INT
        and self.weight_tensor_config.dtype == TensorDataType.INT
        and self.compute_precision != ComputePrecision.INTEGER
    ):
      raise ValueError(
          'Op execution mode must be SRQ (static range quantization) if both'
          ' activation and weight tensors are quantized!'
      )

  def to_dict(self) -> dict[str, Any]:
    """Converts OpQuantizationConfig to dict."""
    return dataclasses.asdict(
        self,
        dict_factory=lambda x: {  # pylint: disable=g-long-lambda
            k: v
            for (k, v) in x
            # Skip None and empty dict values.
            if v is not None and not (isinstance(v, dict) and not v)
        },
    )

  @classmethod
  def from_dict(cls, params: dict[str, Any]) -> 'OpQuantizationConfig':
    """Converts a given dict to OpQuantizationConfig."""
    params_copy = copy.deepcopy(params)
    params_copy['weight_tensor_config'] = TensorQuantizationConfig.from_dict(
        params_copy['weight_tensor_config']
    )
    if 'activation_tensor_config' in params_copy:
      params_copy['activation_tensor_config'] = (
          TensorQuantizationConfig.from_dict(
              params_copy['activation_tensor_config']
          )
      )
    return cls(**params_copy)


@dataclasses.dataclass(frozen=True)
class GraphInfo:
  """Aggregates graph information needed to perform quantization for an op.

  Attributes:
    subgraph_tensors: Tensors in the subgraph.
    buffers: Buffers in the subgraph.
  """

  subgraph_tensors: list[Any]
  buffers: list[Any]


@dataclasses.dataclass(frozen=True)
class OpInfo:
  """Aggregates op information needed to perform quantization for an op.

  Attributes:
    op: The op to be quantized.
    op_name: The name of the op.
    subgraph_op_index: The position of the op in the subgraph.
    op_quant_config: The quantization configuration for the op.
  """

  op: Any
  op_name: TFLOperationName
  subgraph_op_index: int  # Position of the op in the subgraph.
  op_quant_config: OpQuantizationConfig


# Data classes used by model modifier.


# TODO: b/335530570 - This needs to support more than one parameters.
@dataclasses.dataclass
class TransformationInst:
  """Transformation instruction for a tensor.

  Attributes:
    transformation: The transformation to be applied to the tensor.
    tensor_id: The id of the tensor.
    producer: The id of the producer op.
    consumers: The ids of the consumer ops.
    parameters: The quantization parameters for the tensor.
  """

  transformation: QuantTransformation
  tensor_id: int
  producer: Optional[int]
  consumers: list[int]
  parameters: Union[None, UniformQuantParams, NonLinearQuantParams] = None


@dataclasses.dataclass
class TensorTransformationInsts:
  """Transformation instructions for a tensor.

  Attributes:
    tensor_name: The name of the tensor.
    subgraph_id: The id of the subgraph.
    instructions: The transformation instructions for the tensor.
  """

  tensor_name: str
  subgraph_id: int
  instructions: Optional[list[TransformationInst]]


@dataclasses.dataclass(frozen=True)
class TransformationInfo:
  """Transformation information for an op.

  Attributes:
    op_id: The id where op replacement/insertion begins.
    num_ops_added: The number of ops added during the transformation.
    output_tensor_id: The id of the output tensor.
  """

  op_id: int
  num_ops_added: int
  output_tensor_id: int


# Policy is represented as a dict to check the op quantization config.
# Normally the policy is loaded from a json file.
ConfigCheckPolicyDict = collections.OrderedDict[
    TFLOperationName, list[OpQuantizationConfig]
]


def _compare_array_or_none(
    obj1: Optional[np.ndarray], obj2: Optional[np.ndarray]
):
  """Compares two arrays or None.

  Args:
    obj1: The first object to compare.
    obj2: The second object to compare.

  Returns:
    True if both objects are None or both objects are equal.
  """
  if obj1 is None and obj2 is None:
    return True  # Both None, so they're equal.
  elif obj1 is None or obj2 is None:
    return False  # Only one is None, so they're different.
  else:
    return np.array_equal(obj1, obj2)


@dataclasses.dataclass(frozen=True)
class IOOperator:
  """IOOperator class to represent the input and output for a subgraph.

  Attributes:
    inputs: The input tensor ids of the op.
    outputs: The output tensor ids of the op.
    op_key: The op key of the op (input or output).
  """

  inputs: list[int]
  outputs: list[int]
  op_key: TFLOperationName


# The function signature for `get_tensor_quant_params_fn`.
GetTensorQuantParamsFuncSignature = Callable[
    [
        OpInfo,  # op_info
        TensorQuantizationConfig,  # tensor_quant_config
        Optional[np.ndarray],  # tensor_data
        Optional[dict[str, Any]],  # tensor qsv
    ],
    UniformQuantParams,
]



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""flatbuffer utils for the Quantizer."""

import logging
import pathlib
from typing import Any, Optional, Union

import immutabledict
import numpy as np

import os
from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_litert import schema_py_generated as schema  # pylint:disable=g-direct-tensorflow-import


_TFLOpName = qtyping.TFLOperationName

Path = Union[str, pathlib.Path]

TFL_OP_NAME_TO_CODE = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: schema.BuiltinOperator.FULLY_CONNECTED,
    _TFLOpName.BATCH_MATMUL: schema.BuiltinOperator.BATCH_MATMUL,
    _TFLOpName.CONV_2D: schema.BuiltinOperator.CONV_2D,
    _TFLOpName.DEPTHWISE_CONV_2D: schema.BuiltinOperator.DEPTHWISE_CONV_2D,
    _TFLOpName.CONV_2D_TRANSPOSE: schema.BuiltinOperator.TRANSPOSE_CONV,
    _TFLOpName.EMBEDDING_LOOKUP: schema.BuiltinOperator.EMBEDDING_LOOKUP,
    _TFLOpName.SOFTMAX: schema.BuiltinOperator.SOFTMAX,
    _TFLOpName.AVERAGE_POOL_2D: schema.BuiltinOperator.AVERAGE_POOL_2D,
    _TFLOpName.RESHAPE: schema.BuiltinOperator.RESHAPE,
    _TFLOpName.TANH: schema.BuiltinOperator.TANH,
    _TFLOpName.TRANSPOSE: schema.BuiltinOperator.TRANSPOSE,
    _TFLOpName.GELU: schema.BuiltinOperator.GELU,
    _TFLOpName.ADD: schema.BuiltinOperator.ADD,
    _TFLOpName.SUB: schema.BuiltinOperator.SUB,
    _TFLOpName.MUL: schema.BuiltinOperator.MUL,
    _TFLOpName.MEAN: schema.BuiltinOperator.MEAN,
    _TFLOpName.RSQRT: schema.BuiltinOperator.RSQRT,
    _TFLOpName.CONCATENATION: schema.BuiltinOperator.CONCATENATION,
    _TFLOpName.STRIDED_SLICE: schema.BuiltinOperator.STRIDED_SLICE,
    _TFLOpName.SPLIT: schema.BuiltinOperator.SPLIT,
    _TFLOpName.LOGISTIC: schema.BuiltinOperator.LOGISTIC,
    _TFLOpName.SLICE: schema.BuiltinOperator.SLICE,
    _TFLOpName.SUM: schema.BuiltinOperator.SUM,
    _TFLOpName.SELECT: schema.BuiltinOperator.SELECT,
    _TFLOpName.SELECT_V2: schema.BuiltinOperator.SELECT_V2,
    _TFLOpName.STABLEHLO_COMPOSITE: schema.BuiltinOperator.STABLEHLO_COMPOSITE,
    _TFLOpName.DYNAMIC_UPDATE_SLICE: (
        schema.BuiltinOperator.DYNAMIC_UPDATE_SLICE
    ),
    _TFLOpName.PAD: schema.BuiltinOperator.PAD,
    _TFLOpName.SQUARED_DIFFERENCE: schema.BuiltinOperator.SQUARED_DIFFERENCE,
    _TFLOpName.MAX_POOL_2D: schema.BuiltinOperator.MAX_POOL_2D,
    _TFLOpName.RESIZE_BILINEAR: schema.BuiltinOperator.RESIZE_BILINEAR,
    _TFLOpName.RESIZE_NEAREST_NEIGHBOR: (
        schema.BuiltinOperator.RESIZE_NEAREST_NEIGHBOR
    ),
    _TFLOpName.GATHER_ND: schema.BuiltinOperator.GATHER_ND,
    _TFLOpName.PACK: schema.BuiltinOperator.PACK,
    _TFLOpName.UNPACK: schema.BuiltinOperator.UNPACK,
    _TFLOpName.DIV: schema.BuiltinOperator.DIV,
    _TFLOpName.BROADCAST_TO: schema.BuiltinOperator.BROADCAST_TO,
    _TFLOpName.SQRT: schema.BuiltinOperator.SQRT,
    _TFLOpName.GATHER: schema.BuiltinOperator.GATHER,
    _TFLOpName.HARD_SWISH: schema.BuiltinOperator.HARD_SWISH,
    _TFLOpName.MAXIMUM: schema.BuiltinOperator.MAXIMUM,
    _TFLOpName.PADV2: schema.BuiltinOperator.PADV2,
    _TFLOpName.REDUCE_MIN: schema.BuiltinOperator.REDUCE_MIN,
    _TFLOpName.EQUAL: schema.BuiltinOperator.EQUAL,
    _TFLOpName.NOT_EQUAL: schema.BuiltinOperator.NOT_EQUAL,
    _TFLOpName.MIRROR_PAD: schema.BuiltinOperator.MIRROR_PAD,
    _TFLOpName.SPACE_TO_DEPTH: schema.BuiltinOperator.SPACE_TO_DEPTH,
    _TFLOpName.RELU: schema.BuiltinOperator.RELU,
})

TFL_OP_CODE_TO_NAME = immutabledict.immutabledict(
    dict((reversed(item) for item in TFL_OP_NAME_TO_CODE.items()))
)

# Quantized dimension for per-channel quantization.
# See https://www.tensorflow.org/lite/performance/quantization_spec.
TFL_OP_TO_WEIGHT_QUANTIZED_DIM = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: 0,
    _TFLOpName.DEPTHWISE_CONV_2D: 3,
    _TFLOpName.CONV_2D: 0,
    _TFLOpName.EMBEDDING_LOOKUP: 0,
    _TFLOpName.CONV_2D_TRANSPOSE: 0,
})

TFL_OP_TO_BLOCKWISE_WEIGHT_QUANTIZED_DIM = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: 1,
    _TFLOpName.EMBEDDING_LOOKUP: 1,
})

NUM_TFL_DATATYPES = 18
TENSOR_CODE_TO_TYPE = {}
for dtype_code in range(NUM_TFL_DATATYPES):
  TENSOR_CODE_TO_TYPE[dtype_code] = flatbuffer_utils.type_to_name(dtype_code)
TENSOR_CODE_TO_TYPE = immutabledict.immutabledict(TENSOR_CODE_TO_TYPE)
TENSOR_TYPE_TO_CODE = immutabledict.immutabledict(  # pytype: disable=wrong-arg-types
    (reversed(item) for item in TENSOR_CODE_TO_TYPE.items())
)

# Expose functions in litert.python.tools.flatbuffer_utils
write_model = flatbuffer_utils.write_model


def read_model(tflite_model: Union[Path, bytearray]) -> Any:
  """Read and convert the TFLite model into a flatbuffer object.

  Args:
    tflite_model: TFLite model path or bytearray.

  Raises:
    ValueError: Unsupported tflite_model type.

  Returns:
    flatbuffer_model: the flatbuffer_model.
  """
  if isinstance(tflite_model, (str, pathlib.Path)):
    return flatbuffer_utils.read_model(tflite_model)
  elif isinstance(tflite_model, bytes) or isinstance(tflite_model, bytearray):
    return flatbuffer_utils.read_model_from_bytearray(tflite_model)
  else:
    raise ValueError(
        "Unsupported tflite_model type: %s" % type(tflite_model).__name__
    )


def get_model_content(tflite_path: Path) -> bytes:
  """Get the model content (bytes) from the path.

  Args:
    tflite_path: Path to the .tflite.

  Returns:
    The model bytes.
  """
  with open(tflite_path, "rb") as tflite_file:
    return tflite_file.read()


def get_model_buffer(tflite_path: Path) -> bytearray:
  """Get the model buffer from the path.

  Args:
    tflite_path: path to the .tflite.

  Returns:
    model_buffer: the model buffer.
  """
  with open(tflite_path, "rb") as tflite_file:
    return bytearray(tflite_file.read())


def parse_op_tensors(op: Any, subgraph_tensors: list[Any]) -> list[Any]:
  """Parse the op tensors.

  Args:
    op: the op that need to be parsed.
    subgraph_tensors: list of tensors in the subgraph.

  Returns:
    tensors: list of tensors that are associated with the op.
  """

  tensors = []
  for tensor_idx in list(op.outputs) + list(op.inputs):
    if tensor_idx != -1:
      tensors.append(subgraph_tensors[tensor_idx])
  return tensors


def parse_fc_bmm_conv_tensors(
    op: Any,
    subgraph_tensors: list[Any],
    input_index: int = 0,
    weight_index: int = 1,
    bias_index: int = 2,
    output_index: int = 0,
) -> tuple[Any, Any, Any, Any]:
  """Parse tensors in FullyConnected, BatchMatmul, and Convolutions.

  Args:
    op: the TFLite op, must be fully_connected, batch_matmul, or convolution.
    subgraph_tensors: tensors in the subgraph.
    input_index: index for the input tensor.
    weight_index: index for the weight tensor.
    bias_index: index for the bias tensor.
    output_index: index for the output tensor.

  Returns:
    input_tensor, weight_tensor, bias_tensor, output_tensor
  """

  input_tensor = subgraph_tensors[op.inputs[input_index]]
  weight_tensor = subgraph_tensors[op.inputs[weight_index]]
  bias_tensor = None
  if bias_index < len(op.inputs) and op.inputs[bias_index] != -1:
    bias_tensor = subgraph_tensors[op.inputs[bias_index]]
  output_tensor = subgraph_tensors[op.outputs[output_index]]
  return input_tensor, weight_tensor, bias_tensor, output_tensor


# flatbuffer_model has Any type since litert.python.tools.flatbuffer_utils
# is not type annotated.
def buffer_to_tensors(flatbuffer_model: Any) -> dict[int, list[Any]]:
  """Returns a map from buffer id to tensors that use it."""
  buffer_to_tensor_map = {}
  for subgraph in flatbuffer_model.subgraphs:
    for op in subgraph.operators:
      for tensor in parse_op_tensors(op, subgraph.tensors):
        if tensor.buffer not in buffer_to_tensor_map:
          buffer_to_tensor_map[tensor.buffer] = []
        if tensor not in buffer_to_tensor_map[tensor.buffer]:
          buffer_to_tensor_map[tensor.buffer].append(tensor)
  return buffer_to_tensor_map


def get_tensor_name(tensor: Any) -> str:
  """Get the tensor name for a fb tensor.

  Args:
    tensor: tensor in flatbuffer.

  Returns:
    tensor_name: name of the buffer
  """
  return tensor.name.decode("utf-8")


def get_tensor_data(tensor: Any, buffers: list[Any]) -> Optional[np.ndarray]:
  """Get the tensor data.

  Args:
    tensor: tensor in flatbuffer.
    buffers: list of buffers

  Returns:
    tensor_data: data inside the tensor
  """
  tensor_buffer = buffers[tensor.buffer]
  buffer_data = tensor_buffer.data
  if buffer_data is None:
    return None
  data = np.frombuffer(
      buffer_data, dtype=TENSOR_CODE_TO_TYPE[tensor.type].lower()
  )
  if tensor.shape is not None:
    data = np.reshape(data, tensor.shape)
  return data


def has_same_quantization(tensor1: Any, tensor2: Any) -> bool:
  """Check if two tensors have the same quantization.

  Args:
    tensor1: tensor in flatbuffer.
    tensor2: tensor in flatbuffer.

  Returns:
    True if two tensors have the same quantization.
  """

  def to_tuple(val):
    if val is None:
      val = []
    return tuple(val)

  same_type = tensor1.type == tensor2.type

  # Return True if both tensors are not quantized.
  if tensor1.quantization is None and tensor2.quantization is None:
    return True
  if tensor1.quantization.scale is None and tensor2.quantization.scale is None:
    return True

  same_scale = to_tuple(tensor1.quantization.scale) == to_tuple(
      tensor2.quantization.scale
  )
  same_zero_point = to_tuple(tensor1.quantization.zeroPoint) == to_tuple(
      tensor2.quantization.zeroPoint
  )
  same_quantized_dimension = (
      tensor1.quantization.quantizedDimension
      == tensor2.quantization.quantizedDimension
  )
  return (
      same_type and same_scale and same_zero_point and same_quantized_dimension
  )


def is_float_model(flatbuffer_model: Any) -> bool:
  """Checks that the model is float and not already quantized."""
  for subgraph in flatbuffer_model.subgraphs:
    for tensor in subgraph.tensors:
      if tensor.quantization is None:
        continue
      if tensor.quantization.scale is not None:
        return False
  return True


def get_subgraph_input_output_operators(
    subgraph: Any,
) -> list[qtyping.IOOperator]:
  """Get the input/output operators for the subgraph.

  Args:
    subgraph: The subgraph object.

  Returns:
    Input and output operators for the subgraph.
  """
  input_op = qtyping.IOOperator(
      inputs=[],
      outputs=subgraph.inputs,
      op_key=qtyping.TFLOperationName.INPUT,
  )
  output_op = qtyping.IOOperator(
      inputs=subgraph.outputs,
      outputs=[],
      op_key=qtyping.TFLOperationName.OUTPUT,
  )
  return [input_op, output_op]


def get_op_side_effect_subgraphs(
    op: Union[schema.Operator, schema.OperatorT],
) -> list[int]:
  """Get indices of any subgraphs invoked as a side effect of the operator.

  Args:
    op: The operator object.

  Returns:
    A list of subgraph indices invoked by the operator. Empty if the operator
    does not invoke any subgraphs.
  """
  if opts := flatbuffer_utils.get_options_as(
      op, schema.StableHLOCompositeOptionsT
  ):
    return [opts.decompositionSubgraphIndex]
  # Can add other nested ops here (control flow ops, etc).
  return []


def get_op_name_by_index(
    flatbuffer_model: Any, subgraph_id: int, op_index: int
) -> str:
  """Get the op name from the flatbuffer model."""
  op = flatbuffer_model.subgraphs[subgraph_id].operators[op_index]
  builtin_code = flatbuffer_model.operatorCodes[op.opcodeIndex].builtinCode
  return TFL_OP_CODE_TO_NAME[builtin_code]


def get_op_scope(
    op: Any, subgraph_tensors: list[Any], max_length: int = 10000
) -> str:
  """Get the op scope.

  Op scope is defined by the output tensor names (following ModelExplorer). If
  no output tensors are present, the input tensor names are used instead.

  Args:
    op: The op that needs to be parsed.
    subgraph_tensors: Tensors in the subgraph.
    max_length: The maximum length of the scope string. If the scope string is
      longer than this length, it will be truncated to this length to avoid
      overwhelming regex matching engine.

  Returns:
    Scope for the op.
  """

  def _get_valid_tensor_names(tensor_indices: list[int]) -> list[str]:
    """Gets names of tensors for valid indices."""
    names = []
    for idx in tensor_indices:
      if idx != -1:
        names.append(get_tensor_name(subgraph_tensors[idx]))
    return [name for name in names if name]

  # Op scope is determined by output tensors.
  tensor_names = _get_valid_tensor_names(op.outputs)
  # If no output tensors, use input tensors.
  if not tensor_names:
    tensor_names = _get_valid_tensor_names(op.inputs)

  scope = ";".join(tensor_names)
  # Add a trailing semicolon to help distinguish scopes with the same prefix.
  if tensor_names:
    scope += ";"
  if len(scope) > max_length:
    logging.warning(
        "Op scope is too long, truncating to %d characters. Truncated scope:"
        " %s",
        max_length,
        scope,
    )
  return scope[:max_length]
